import unittest

import numpy as np
import pinocchio as pin
from test_case import PinocchioTestCase


class TestFrameBindings(PinocchioTestCase):
    def setUp(self):
        self.model = pin.buildSampleModelHumanoidRandom()
        self.parent_idx = (
            self.model.getJointId("rarm2_joint")
            if self.model.existJointName("rarm2_joint")
            else (self.model.njoints - 1)
        )
        self.frame_name = self.model.names[self.parent_idx] + "_frame"
        self.frame_placement = pin.SE3.Random()
        self.frame_type = pin.FrameType.OP_FRAME
        self.model.addFrame(
            pin.Frame(
                self.frame_name,
                self.parent_idx,
                0,
                self.frame_placement,
                self.frame_type,
            )
        )
        self.frame_idx = self.model.getFrameId(self.frame_name)

        self.data = self.model.createData()
        self.q = pin.randomConfiguration(self.model)
        self.v = np.random.rand(self.model.nv)
        self.a = np.random.rand(self.model.nv)

    def tearDown(self):
        del self.model

    def test_derivatives(self):
        model = self.model
        data = self.data

        q = self.q
        v = self.v
        a = self.a

        pin.computeForwardKinematicsDerivatives(model, data, q, v, a)

        pin.getFrameVelocityDerivatives(model, data, self.frame_idx, pin.WORLD)
        pin.getFrameVelocityDerivatives(model, data, self.frame_idx, pin.LOCAL)
        pin.getFrameVelocityDerivatives(
            model, data, self.frame_idx, pin.LOCAL_WORLD_ALIGNED
        )

        pin.getFrameAccelerationDerivatives(model, data, self.frame_idx, pin.WORLD)
        pin.getFrameAccelerationDerivatives(model, data, self.frame_idx, pin.LOCAL)
        pin.getFrameAccelerationDerivatives(
            model, data, self.frame_idx, pin.LOCAL_WORLD_ALIGNED
        )


if __name__ == "__main__":
    unittest.main()



# Parse input arguments
import argparse
import math
import sys
import time

import hppfcl as fcl
import numpy as np
import pinocchio as pin
from pinocchio.visualize import MeshcatVisualizer as Visualizer

parser = argparse.ArgumentParser()
parser.add_argument(
    "--with-cart",
    help="Add a cart at the base of the pendulum to simulate a cart pole system.",
    action="store_true",
)
parser.add_argument(
    "-N", help="Number of pendulums compositing the dynamical system.", type=int
)
args = parser.parse_args()

if args.N:
    N = args.N
else:
    N = 1  # number of pendulums

model = pin.Model()
geom_model = pin.GeometryModel()

parent_id = 0

if args.with_cart:
    cart_radius = 0.1
    cart_length = 5 * cart_radius
    cart_mass = 2.0
    joint_name = "joint_cart"

    geometry_placement = pin.SE3.Identity()
    geometry_placement.rotation = pin.Quaternion(
        np.array([0.0, 0.0, 1.0]), np.array([0.0, 1.0, 0.0])
    ).toRotationMatrix()

    joint_id = model.addJoint(
        parent_id, pin.JointModelPY(), pin.SE3.Identity(), joint_name
    )

    body_inertia = pin.Inertia.FromCylinder(cart_mass, cart_radius, cart_length)
    body_placement = geometry_placement
    # We need to rotate the inertia as it is expressed in the LOCAL frame of the
    # geometry.
    model.appendBodyToJoint(joint_id, body_inertia, body_placement)

    shape_cart = fcl.Cylinder(cart_radius, cart_length)

    geom_cart = pin.GeometryObject(
        "shape_cart", joint_id, geometry_placement, shape_cart
    )
    geom_cart.meshColor = np.array([1.0, 0.1, 0.1, 1.0])
    geom_model.addGeometryObject(geom_cart)

    parent_id = joint_id
else:
    base_radius = 0.2
    shape_base = fcl.Sphere(base_radius)
    geom_base = pin.GeometryObject("base", 0, pin.SE3.Identity(), shape_base)
    geom_base.meshColor = np.array([1.0, 0.1, 0.1, 1.0])
    geom_model.addGeometryObject(geom_base)

joint_placement = pin.SE3.Identity()
body_mass = 1.0
body_radius = 0.1

for k in range(N):
    joint_name = "joint_" + str(k + 1)
    joint_id = model.addJoint(
        parent_id, pin.JointModelRX(), joint_placement, joint_name
    )

    body_inertia = pin.Inertia.FromSphere(body_mass, body_radius)
    body_placement = joint_placement.copy()
    body_placement.translation[2] = 1.0
    model.appendBodyToJoint(joint_id, body_inertia, body_placement)

    geom1_name = "ball_" + str(k + 1)
    shape1 = fcl.Sphere(body_radius)
    geom1_obj = pin.GeometryObject(geom1_name, joint_id, body_placement, shape1)
    geom1_obj.meshColor = np.ones(4)
    geom_model.addGeometryObject(geom1_obj)

    geom2_name = "bar_" + str(k + 1)
    shape2 = fcl.Cylinder(body_radius / 4.0, body_placement.translation[2])
    shape2_placement = body_placement.copy()
    shape2_placement.translation[2] /= 2.0

    geom2_obj = pin.GeometryObject(geom2_name, joint_id, shape2_placement, shape2)
    geom2_obj.meshColor = np.array([0.0, 0.0, 0.0, 1.0])
    geom_model.addGeometryObject(geom2_obj)

    parent_id = joint_id
    joint_placement = body_placement.copy()


visual_model = geom_model
# Initialize the viewer.
try:
    viz = Visualizer(model, geom_model, visual_model)
    viz.initViewer()
except ImportError as err:
    print(
        "Error while initializing the viewer. "
        "It seems you should install gepetto-viewer"
    )
    print(err)
    sys.exit(0)

try:
    viz.loadViewerModel("pinocchio")
except AttributeError as err:
    print(
        "Error while loading the viewer model. "
        "It seems you should start gepetto-viewer"
    )
    print(err)
    sys.exit(0)

# Display a robot configuration.
q0 = pin.neutral(model)
viz.display(q0)

# Play a bit with the simulation
dt = 0.01
T = 5

N = math.floor(T / dt)

model.lowerPositionLimit.fill(-math.pi)
model.upperPositionLimit.fill(+math.pi)

if args.with_cart:
    model.lowerPositionLimit[0] = model.upperPositionLimit[0] = 0.0

data_sim = model.createData()

t = 0.0
q = pin.randomConfiguration(model)
v = np.zeros(model.nv)
tau_control = np.zeros(model.nv)
damping_value = 0.1
for k in range(N):
    tic = time.time()
    tau_control = -damping_value * v  # small damping
    a = pin.aba(model, data_sim, q, v, tau_control)  # Forward dynamics

    # Semi-explicit integration
    v += a * dt
    q = pin.integrate(model, q, v * dt)  # Configuration integration

    viz.display(q)
    toc = time.time()
    ellapsed = toc - tic

    dt_sleep = max(0, dt - (ellapsed))
    time.sleep(dt_sleep)
    t += dt



#
# In this short script, we show how to use RobotWrapper
# integrating different kinds of viewers
#

import os
from pathlib import Path
from sys import argv

import pinocchio as pin
from pinocchio.robot_wrapper import RobotWrapper
from pinocchio.visualize import GepettoVisualizer, MeshcatVisualizer

# If you want to visualize the robot in this example,
# you can choose which visualizer to employ
# by specifying an option from the command line:
# GepettoVisualizer: -g
# MeshcatVisualizer: -m
VISUALIZER = None
if len(argv) > 1:
    opt = argv[1]
    if opt == "-g":
        VISUALIZER = GepettoVisualizer
    elif opt == "-m":
        VISUALIZER = MeshcatVisualizer
    else:
        raise ValueError("Unrecognized option: " + opt)

# Load the URDF model with RobotWrapper
# Conversion with str seems to be necessary when executing this file with ipython
model_path = Path(os.environ.get("EXAMPLE_ROBOT_DATA_MODEL_DIR"))
mesh_dir = model_path.parent.parent
urdf_filename = "talos_reduced.urdf"
urdf_model_path = model_path / "talos_data/robots" / urdf_filename

robot = RobotWrapper.BuildFromURDF(urdf_model_path, mesh_dir, pin.JointModelFreeFlyer())

# alias
model = robot.model
data = robot.data

# do whatever, e.g. compute the center of mass position expressed in the world frame
q0 = robot.q0
com = robot.com(q0)

# This last command is similar to:
com2 = pin.centerOfMass(model, data, q0)

# Show model with a visualizer of your choice
if VISUALIZER:
    robot.setVisualizer(VISUALIZER())
    robot.initViewer()
    robot.loadViewerModel("pinocchio")
    robot.display(q0)



# This examples shows how to load and move a robot in meshcat.
# Note: this feature requires Meshcat to be installed, this can be done using
# pip install --user meshcat

import os
import sys
from pathlib import Path

import numpy as np
import pinocchio as pin
from pinocchio.visualize import MeshcatVisualizer

# Load the URDF model.
model_path = Path(os.environ.get("EXAMPLE_ROBOT_DATA_MODEL_DIR"))
mesh_dir = model_path.parent.parent
urdf_filename = "romeo_small.urdf"
urdf_model_path = model_path / "romeo_description/urdf" / urdf_filename

model, collision_model, visual_model = pin.buildModelsFromUrdf(
    urdf_model_path, mesh_dir, pin.JointModelFreeFlyer()
)

# Start a new MeshCat server and client.
# Note: the server can also be started separately using the "meshcat-server" command in
# a terminal:
# this enables the server to remain active after the current script ends.
#
# Option open=True pens the visualizer.
# Note: the visualizer can also be opened seperately by visiting the provided URL.
try:
    viz = MeshcatVisualizer(model, collision_model, visual_model)
    viz.initViewer(open=True)
except ImportError as err:
    print(
        "Error while initializing the viewer. "
        "It seems you should install Python meshcat"
    )
    print(err)
    sys.exit(0)

# Load the robot in the viewer.
# Color is needed here because the Romeo URDF doesn't contain any color, so the default
# color results in an invisible robot (alpha value set to 0).
viz.loadViewerModel(color=[0.0, 0.0, 0.0, 1.0])

# Display a robot configuration.
q0 = np.array(
    [
        0,
        0,
        0.840252,
        0,
        0,
        0,
        1,  # Free flyer
        0,
        0,
        -0.3490658,
        0.6981317,
        -0.3490658,
        0,  # left leg
        0,
        0,
        -0.3490658,
        0.6981317,
        -0.3490658,
        0,  # right leg
        0,  # chest
        1.5,
        0.6,
        -0.5,
        -1.05,
        -0.4,
        -0.3,
        -0.2,  # left arm
        0,
        0,
        0,
        0,  # head
        1.5,
        -0.6,
        0.5,
        1.05,
        -0.4,
        -0.3,
        -0.2,  # right arm
    ]
).T
viz.display(q0)

# Display another robot.
red_robot_viz = MeshcatVisualizer(model, collision_model, visual_model)
red_robot_viz.initViewer(viz.viewer)
red_robot_viz.loadViewerModel(rootNodeName="red_robot", color=[1.0, 0.0, 0.0, 0.5])
q = q0.copy()
q[1] = 1.0
red_robot_viz.display(q)



import unittest

import numpy as np
import pinocchio as pin
from test_case import PinocchioTestCase as TestCase


class TestDeriavtives(TestCase):
    def setUp(self):
        self.model = pin.buildSampleModelHumanoidRandom()
        self.data = self.model.createData()

        qmax = np.full((self.model.nq, 1), np.pi)
        self.q = pin.randomConfiguration(self.model, -qmax, qmax)
        self.v = np.random.rand(self.model.nv)
        self.a = np.random.rand(self.model.nv)

    def test_centroidal_derivatives(self):
        res = pin.computeCentroidalDynamicsDerivatives(
            self.model, self.data, self.q, self.v, self.a
        )

        self.assertTrue(len(res) == 4)

        data2 = self.model.createData()
        pin.computeCentroidalMomentumTimeVariation(
            self.model, data2, self.q, self.v, self.a
        )

        self.assertApprox(self.data.hg, data2.hg)
        self.assertApprox(self.data.dhg, data2.dhg)

        data3 = self.model.createData()
        pin.computeRNEADerivatives(self.model, data3, self.q, self.v, self.a)
        res2 = pin.getCentroidalDynamicsDerivatives(self.model, data3)

        for k in range(4):
            self.assertApprox(res[k], res2[k])


if __name__ == "__main__":
    unittest.main()



"""
Example of optimal control resolution by direct optimization of a single trajectory.
"""

import signal
import time

import matplotlib.pyplot as plt
import numpy as np
from pendulum import Pendulum
from scipy.optimize import fmin_l_bfgs_b

env = Pendulum(1)
NSTEPS = 50
x0 = env.reset().copy()


def cost(U):
    """Cost for a trajectory starting at state X0 with control U"""
    env.reset(x0)
    csum = 0.0
    for t in range(NSTEPS):
        u = U[env.nu * t : env.nu * (t + 1)]  # Control at time step <t>
        _, r = env.step(u)  # Pendulum step, with reward r
        csum += r  # Cumulative sum
    return -csum  # Returns cost ie negative reward


def display(U, verbose=False):
    """Display the trajectory on Gepetto viewer."""
    x = x0.copy()
    if verbose:
        print("U = ", " ".join(map(lambda u: f"{u:.1f}", np.asarray(U).flatten())))
    for i in range(len(U) / env.nu):
        env.dynamics(x, U[env.nu * i : env.nu * (i + 1)], True)
        env.display(x)
        time.sleep(5e-2)
        if verbose:
            print(f"X{i}")


class CallBack:
    """Call back function used to follow optimizer steps."""

    def __init__(self):
        self.iter = 0
        self.withdisplay = False
        self.h_rwd = []

    def __call__(self, U):
        print(
            "Iteration ",
            self.iter,
            " ".join(map(lambda u: f"{u:.1f}", np.asarray(U).flatten())),
        )
        self.iter += 1
        self.U = U.copy()
        self.h_rwd.append(cost(U))
        if self.withdisplay:
            display(U)  # Display if CTRL-Z has been pressed.

    def setWithDisplay(self, boolean=None):
        self.withdisplay = not self.withdisplay if boolean is None else boolean


callback = CallBack()
signal.signal(signal.SIGTSTP, lambda x, y: callback.setWithDisplay())

# --- OCP resolution
# Initial guess for the control trajectory.
U0 = np.zeros(NSTEPS * env.nu) - env.umax
bounds = (
    [
        [-env.umax, env.umax],
    ]
    * env.nu
    * NSTEPS
)  # Set control bounds to environment umax.

# Start BFGS optimization routine
U, c, info = fmin_l_bfgs_b(
    cost, x0=U0, callback=callback, approx_grad=True, bounds=bounds
)

# When done, display the trajectory in Gepetto-viewer
display(U, True)

plt.plot(callback.h_rwd)
plt.show()



import os
import unittest
from pathlib import Path

import pinocchio as pin


@unittest.skipUnless(pin.WITH_URDFDOM, "Needs URDFDOM")
class TestBuildGeomFromUrdfMemoryCheck(unittest.TestCase):
    def setUp(self):
        self.model_dir = Path(os.environ.get("EXAMPLE_ROBOT_DATA_MODEL_DIR"))
        self.model_path = self.model_dir / "ur_description/urdf/ur5_robot.urdf"

    def test_load(self):
        model = pin.buildModelFromUrdf(self.model_path)
        for _ in range(2):
            pin.buildGeomFromUrdf(
                model,
                self.model_path,
                pin.COLLISION,
                package_dirs=self.model_dir.parent.parent,
            )


if __name__ == "__main__":
    unittest.main()



import unittest

import pinocchio as pin
from pinocchio.utils import zero
from test_case import PinocchioTestCase as TestCase


class TestRegressorBindings(TestCase):
    def test_staticRegressor(self):
        model = pin.buildSampleModelHumanoidRandom()

        data = model.createData()
        data_ref = model.createData()

        model.lowerPositionLimit[:7] = -1.0
        model.upperPositionLimit[:7] = 1.0

        q = pin.randomConfiguration(model)
        pin.computeStaticRegressor(model, data, q)

        phi = zero(4 * (model.njoints - 1))
        for k in range(1, model.njoints):
            Y = model.inertias[k]
            phi[4 * (k - 1)] = Y.mass
            phi[4 * k - 3 : 4 * k] = Y.mass * Y.lever

        static_com_ref = pin.centerOfMass(model, data_ref, q)
        static_com = data.staticRegressor.dot(phi)

        self.assertApprox(static_com, static_com_ref)

    def test_bodyRegressor(self):
        In = pin.Inertia.Random()
        v = pin.Motion.Random()
        a = pin.Motion.Random()

        f = In * a + In.vxiv(v)

        f_regressor = pin.bodyRegressor(v, a).dot(In.toDynamicParameters())

        self.assertApprox(f_regressor, f.vector)

    def test_jointBodyRegressor(self):
        model = pin.buildSampleModelManipulator()
        data = model.createData()

        JOINT_ID = model.njoints - 1

        q = pin.randomConfiguration(model)
        v = pin.utils.rand(model.nv)
        a = pin.utils.rand(model.nv)

        pin.rnea(model, data, q, v, a)

        f = data.f[JOINT_ID]

        f_regressor = pin.jointBodyRegressor(model, data, JOINT_ID).dot(
            model.inertias[JOINT_ID].toDynamicParameters()
        )

        self.assertApprox(f_regressor, f.vector)

    def test_frameBodyRegressor(self):
        model = pin.buildSampleModelManipulator()

        JOINT_ID = model.njoints - 1

        framePlacement = pin.SE3.Random()
        FRAME_ID = model.addBodyFrame("test_body", JOINT_ID, framePlacement, -1)

        data = model.createData()

        q = pin.randomConfiguration(model)
        v = pin.utils.rand(model.nv)
        a = pin.utils.rand(model.nv)

        pin.rnea(model, data, q, v, a)

        f = framePlacement.actInv(data.f[JOINT_ID])
        In = framePlacement.actInv(model.inertias[JOINT_ID])

        f_regressor = pin.frameBodyRegressor(model, data, FRAME_ID).dot(
            In.toDynamicParameters()
        )

        self.assertApprox(f_regressor, f.vector)

    def test_joint_torque_regressor(self):
        model = pin.buildSampleModelHumanoidRandom()
        model.lowerPositionLimit[:7] = -1.0
        model.upperPositionLimit[:7] = 1.0

        data = model.createData()
        data_ref = model.createData()

        q = pin.randomConfiguration(model)
        v = pin.utils.rand(model.nv)
        a = pin.utils.rand(model.nv)

        pin.rnea(model, data_ref, q, v, a)

        params = zero(10 * (model.njoints - 1))
        for i in range(1, model.njoints):
            params[(i - 1) * 10 : i * 10] = model.inertias[i].toDynamicParameters()

        pin.computeJointTorqueRegressor(model, data, q, v, a)

        tau_regressor = data.jointTorqueRegressor.dot(params)

        self.assertApprox(tau_regressor, data_ref.tau)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import pinocchio as pin
from pinocchio.utils import eye, zero
from test_case import PinocchioTestCase as TestCase


class TestSymmetric3Bindings(TestCase):
    def test_zero_getter(self):
        S = pin.Symmetric3.Zero()
        self.assertTrue(np.allclose(zero(6), S.data))

    def test_identity_getters_matrix_conversion(self):
        S = pin.Symmetric3.Identity()
        id_vec = np.array([1.0, 0.0, 1.0, 0.0, 0.0, 1.0])
        id_matrix = pin.Symmetric3(id_vec).matrix()
        self.assertTrue(np.allclose(eye(3), id_matrix))
        self.assertTrue(np.allclose(np.array([1.0, 0.0, 1.0, 0.0, 0.0, 1.0]), S.data))
        self.assertTrue(np.allclose(eye(3), S.matrix()))

    def test_setRandom_matrix_conversion(self):
        S = pin.Symmetric3.Identity()
        S.setRandom()
        self.assertFalse(np.allclose(eye(3), S.matrix()))
        S_matrix = S.matrix()
        tri_upper = np.triu(S_matrix, k=1)
        tri_lower = np.tril(S_matrix, k=-1)
        self.assertTrue(np.allclose(tri_upper, tri_lower.T))

    def test_setZero(self):
        S = pin.Symmetric3.Zero()
        S.setRandom()
        S.setZero()
        self.assertTrue(np.allclose(zero(6), S.data))

    def test_setIdentity_matrix_conversion(self):
        S = pin.Symmetric3.Zero()
        S.setIdentity()
        self.assertTrue(np.allclose(eye(3), S.matrix()))

    def test_setDiagonal(self):
        S = pin.Symmetric3.Zero()
        S.setDiagonal(np.ones(3) * 2)
        self.assertTrue(np.allclose(eye(3) * 2, S.matrix()))


if __name__ == "__main__":
    unittest.main()



import os
from pathlib import Path

import casadi
import numpy as np
import pinocchio as pin
import pinocchio.casadi as cpin

# Problem parameters
x_goal = [1, 0, 1.5, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0]
x0 = [0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0]
nodes = 80
dt = 0.02

# Quadcopter parameters
d_cog, cf, cm = 0.1525, 6.6e-5, 1e-6

tau_f = np.array(
    [
        [0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0],
        [1.0, 1.0, 1.0, 1.0],
        [0.0, d_cog, 0.0, -d_cog],
        [-d_cog, 0.0, d_cog, 0.0],
        [-cm / cf, cm / cf, -cm / cf, cm / cf],
    ]
)

# Other variables
x_nom = [0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0]


def actuation_model():
    u = casadi.SX.sym("u", 4)  # rotor velocities
    tau = tau_f @ u

    return casadi.Function("act_model", [u], [tau], ["u"], ["tau"])


def state_integrate(model):
    q = casadi.SX.sym("dq", model.nq)
    dq = casadi.SX.sym("q", model.nv)
    v = casadi.SX.sym("v", model.nv)
    dv = casadi.SX.sym("dv", model.nv)

    q_next = cpin.integrate(model, q, dq)
    v_next = v + dv

    dx = casadi.vertcat(dq, dv)
    x = casadi.vertcat(q, v)
    x_next = casadi.vertcat(q_next, v_next)

    return casadi.Function("integrate", [x, dx], [x_next], ["x", "dx"], ["x_next"])


def state_difference(model):
    q0 = casadi.SX.sym("q0", model.nq)
    q1 = casadi.SX.sym("q1", model.nq)
    v0 = casadi.SX.sym("v0", model.nv)
    v1 = casadi.SX.sym("v1", model.nv)

    q_diff = cpin.difference(model, q0, q1)
    v_diff = v1 - v0

    x0 = casadi.vertcat(q0, v0)
    x1 = casadi.vertcat(q1, v1)
    x_diff = casadi.vertcat(q_diff, v_diff)

    return casadi.Function("difference", [x0, x1], [x_diff], ["x0", "x1"], ["x_diff"])


def euler_integration(model, data, dt):
    nu = 4
    u = casadi.SX.sym("u", nu)

    # tau = casadi.vertcat(np.zeros(model.nv - nu), u)
    tau = actuation_model()(u)

    q = casadi.SX.sym("q", model.nq)
    v = casadi.SX.sym("v", model.nv)

    a = cpin.aba(model, data, q, v, tau)

    dq = v * dt + a * dt**2
    dv = a * dt

    x = casadi.vertcat(q, v)
    dx = casadi.vertcat(dq, dv)
    x_next = state_integrate(model)(x, dx)

    return casadi.Function("int_dyn", [x, u], [x_next], ["x", "u"], ["x_next"])


def cost_quadratic_state_error(model):
    dx = casadi.SX.sym("dx", model.nv * 2)

    x_N = state_integrate(model)(x_nom, dx)
    e_goal = state_difference(model)(x_N, x_goal)

    cost = 0.5 * e_goal.T @ e_goal

    return casadi.Function("quad_cost", [dx], [cost], ["dx"], ["cost"])


class OptimalControlProblem:
    def __init__(self, model, terminal_soft_constraint=False):
        self.opti = casadi.Opti()

        self.model = model
        self.data = self.model.createData()

        self.c_model = cpin.Model(self.model)
        self.c_data = self.c_model.createData()

        nv = self.c_model.nv
        nu = 4

        self.c_dxs = self.opti.variable(2 * nv, nodes + 1)  # state trajectory
        self.c_us = self.opti.variable(nu, nodes)  # control trajectory

        # Objective function
        obj = 0

        # State & Control regularization
        for i in range(nodes):
            x_i = state_integrate(self.c_model)(x_nom, self.c_dxs[:, i])
            e_reg = state_difference(self.c_model)(x_nom, x_i)
            obj += (
                1e-5 * 0.5 * e_reg.T @ e_reg
                + 1e-5 * 0.5 * self.c_us[:, i].T @ self.c_us[:, i]
            )
        if terminal_soft_constraint:
            obj += 1000 * cost_quadratic_state_error(self.c_model)(self.c_dxs[:, nodes])

        self.opti.minimize(obj)

        # Dynamical constraints
        for i in range(nodes):
            x_i = state_integrate(self.c_model)(x_nom, self.c_dxs[:, i])
            x_i_1 = state_integrate(self.c_model)(x_nom, self.c_dxs[:, i + 1])
            f_x_u = euler_integration(self.c_model, self.c_data, dt)(
                x_i, self.c_us[:, i]
            )
            gap = state_difference(self.c_model)(f_x_u, x_i_1)

            self.opti.subject_to(gap == [0] * 12)

        # Control constraints
        self.opti.subject_to(self.opti.bounded(0, self.c_us, 5))

        # Final constraint
        if not terminal_soft_constraint:
            x_N = state_integrate(self.c_model)(x_nom, self.c_dxs[:, nodes])
            e_goal = state_difference(self.c_model)(x_N, x_goal)
            self.opti.subject_to(e_goal == [0] * 12)

        # Initial state
        x_0 = state_integrate(self.c_model)(x_nom, self.c_dxs[:, 0])
        self.opti.subject_to(state_difference(self.c_model)(x0, x_0) == [0] * 12)

        # Warm start
        self.opti.set_initial(
            self.c_dxs, np.vstack([np.zeros(12) for _ in range(nodes + 1)]).T
        )
        self.opti.set_initial(
            self.c_us, np.vstack([np.zeros(4) for _ in range(nodes)]).T
        )

    def solve(self, approx_hessian=True):
        opts = {"verbose": False}
        opts["ipopt"] = {
            "max_iter": 1000,
            "linear_solver": "mumps",
            "tol": 3.82e-6,
            "mu_strategy": "adaptive",
        }

        if approx_hessian:
            opts["ipopt"]["hessian_approximation"] = "limited-memory"

        # Solver initialization
        self.opti.solver("ipopt", opts)  # set numerical backend

        try:
            self.sol = self.opti.solve()
        except:  # noqa: E722
            self.sol = self.opti.debug

        self._retract_trajectory()
        self._compute_gaps()

    def _retract_trajectory(self):
        self.xs = []
        self.us = []
        self.gaps = []

        nq = self.model.nq
        nv = self.model.nv

        for idx, (dx_sol, u_sol) in enumerate(
            zip(self.sol.value(self.c_dxs).T, self.sol.value(self.c_us).T)
        ):
            q = pin.integrate(self.model, np.array(x_nom)[:nq], dx_sol[:nv])
            v = dx_sol[nv:]

            self.xs.append(np.concatenate([q, v]))
            self.us.append(u_sol)

        q = pin.integrate(
            self.model, np.array(x_nom)[:nq], self.sol.value(self.c_dxs).T[nodes, :nv]
        )
        v = self.sol.value(self.c_dxs).T[nodes, nv:]
        self.xs.append(np.concatenate([q, v]))

    def _compute_gaps(self):
        self.gaps = {"vector": [np.zeros(self.model.nv * 2)], "norm": [0]}

        nq = self.model.nq
        _nv = self.model.nv

        for idx, (x, u) in enumerate(zip(self.xs, self.us)):
            x_pin = self._simulate_step(x, u)

            gap_q = pin.difference(self.model, x_pin[:nq], self.xs[idx + 1][:nq])
            gap_v = self.xs[idx + 1][nq:] - x_pin[nq:]

            gap = np.concatenate([gap_q, gap_v])
            self.gaps["vector"].append(gap)
            self.gaps["norm"].append(np.linalg.norm(gap))

    def _simulate_step(self, x, u):
        nq = self.model.nq
        _nv = self.model.nv

        q = x[:nq]
        v = x[nq:]

        tau = tau_f @ u

        a = pin.aba(self.model, self.data, q, v, tau)

        dq = v * dt + a * dt**2
        dv = a * dt

        q_next = pin.integrate(self.model, q, dq)
        v_next = v + dv

        x_next = np.concatenate([q_next, v_next])

        return x_next


def main():
    model_path = Path(os.environ.get("EXAMPLE_ROBOT_DATA_MODEL_DIR"))
    mesh_dir = model_path.parent.parent
    urdf_filename = model_path / "hector_description/robots/quadrotor_base.urdf"
    model = pin.robot_wrapper.RobotWrapper.BuildFromURDF(
        urdf_filename, mesh_dir, pin.JointModelFreeFlyer()
    ).model

    oc_problem = OptimalControlProblem(model, terminal_soft_constraint=False)

    oc_problem.solve(approx_hessian=True)

    # --------------PLOTS-----------
    try:
        import matplotlib.pyplot as plt

        _, axs0 = plt.subplots(nrows=2)

        xs = np.vstack(oc_problem.xs)
        axs0[0].plot(xs[:, :3])
        axs0[0].set_title("Quadcopter position")

        axs0[1].plot(oc_problem.gaps["norm"])
        axs0[1].set_title("Multiple shooting node gaps")

        _, axs1 = plt.subplots(nrows=4)
        us = np.vstack(oc_problem.us)

        for idx, ax in enumerate(axs1):
            ax.plot(us[:, idx])

        plt.show(block=False)
    except ImportError as err:
        print(
            "Error while initializing the viewer. "
            "It seems you should install Python meshcat"
        )
        raise err


if __name__ == "__main__":
    main()



import numpy as np
import pinocchio as pin
from pinocchio.robot_wrapper import RobotWrapper
from pinocchio.utils import eye


class MobileRobotWrapper(RobotWrapper):
    """
    The class models a mobile robot with UR5 arm, feature 3+6 DOF.
    The configuration of the basis is represented by [X, Y, cos, sin], with the
    additionnal constraint that (cos, sin) should have norm 1. Hence take care when
    sampling and integrating configurations.
    The robot is depicted as an orange box with two black cylinders (wheels) atop of
    which is the classical (realistic) UR5 model.
    The model also features to OPERATIONAL frames, named "mobile" (on the front of the
    mobile basis, 30cm above the ground) and "tool" (at the very end of the effector).
    """

    def __init__(self, urdf, pkgs):
        self.initFromURDF(urdf, pkgs, pin.JointModelPlanar())

        M0 = pin.SE3(eye(3), np.array([0.0, 0.0, 0.6]))
        self.model.jointPlacements[2] = M0 * self.model.jointPlacements[2]
        self.visual_model.geometryObjects[0].placement = (
            M0 * self.visual_model.geometryObjects[0].placement
        )
        self.visual_data.oMg[0] = M0 * self.visual_data.oMg[0]

        # Placement of the "mobile" frame wrt basis center.
        basisMop = pin.SE3(eye(3), np.array([0.3, 0.0, 0.1]))
        self.model.addFrame(pin.Frame("mobile", 1, 1, basisMop, pin.FrameType.OP_FRAME))

        # Placement of the tool frame wrt end effector frame (located at the center of
        # the wrist)
        effMop = pin.SE3(eye(3), np.maarraytrix([0.0, 0.08, 0.095]))
        self.model.addFrame(pin.Frame("tool", 6, 6, effMop, pin.FrameType.OP_FRAME))

        # Create data again after setting frames
        self.data = self.model.createData()

    def initDisplay(self, loadModel):
        RobotWrapper.initDisplay(self, loadModel=loadModel)
        if loadModel and not hasattr(self, "display"):
            RobotWrapper.initDisplay(self, loadModel=False)

        try:
            # self.viewer.gui.deleteNode('world/mobilebasis', True)
            self.viewer.gui.addBox(
                "world/mobilebasis", 0.25, 0.25, 0.25, [0.8, 0.2, 0.2, 1]
            )
            self.viewer.gui.addCylinder(
                "world/mobilewheel1", 0.05, 0.45, [0.1, 0.0, 0.0, 1]
            )
            self.viewer.gui.addCylinder(
                "world/mobilewheel2", 0.05, 0.45, [0.1, 0.0, 0.0, 1]
            )

            self.viewer.gui.setStaticTransform(
                "world/mobilebasis", [0.0, 0.0, 0.35, 1.0, 0.0, 0.0, 0.0]
            )
            self.viewer.gui.setStaticTransform(
                "world/mobilewheel1", [+0.15, 0.0, 0.05, 0.7, 0.7, 0.0, 0.0]
            )
            self.viewer.gui.setStaticTransform(
                "world/mobilewheel2", [-0.15, 0.0, 0.05, 0.7, 0.7, 0.0, 0.0]
            )

            self.viewer.gui.addXYZaxis(
                "world/framebasis", [1.0, 0.0, 0.0, 1.0], 0.03, 0.1
            )
            self.viewer.gui.addXYZaxis(
                "world/frametool", [1.0, 0.0, 0.0, 1.0], 0.03, 0.1
            )
        except:  # noqa: E722
            print("Error when adding 3d objects in the viewer ... ignore")

    def display(self, q):
        RobotWrapper.display(self, q)
        M1 = self.data.oMi[1]
        self.viewer.gui.applyConfiguration(
            "world/mobilebasis", pin.SE3ToXYZQUATtuple(M1)
        )
        self.viewer.gui.applyConfiguration(
            "world/mobilewheel1", pin.SE3ToXYZQUATtuple(M1)
        )
        self.viewer.gui.applyConfiguration(
            "world/mobilewheel2", pin.SE3ToXYZQUATtuple(M1)
        )
        self.viewer.gui.refresh()

        pin.updateFramePlacements(self.model, self.data)
        self.viewer.gui.applyConfiguration(
            "world/framebasis", pin.SE3ToXYZQUATtuple(self.data.oMf[-2])
        )
        self.viewer.gui.applyConfiguration(
            "world/frametool", pin.SE3ToXYZQUATtuple(self.data.oMf[-1])
        )

        self.viewer.gui.refresh()



import numpy as np
import pinocchio as pin

# This example show how to:
#  - Construct a kinematics chain with ModelGraph API
#  - Build a Model with b1 as root body with a fixed base
#  - Build a Model with b3 as root body with a free flyer as root joint
#  - Use ModelConfigurationConverter API to convert configuration and velocity
#    vector from the first model to the second
# Construct kinematics chain with ModelGraph API.
g = pin.graph.ModelGraph()

I_I = pin.Inertia.Identity()
X_I = pin.SE3.Identity()

g.addBody("b1", I_I)
g.addBody("b2", I_I)
g.addBody("b3", I_I)
g.addJoint(
    "j1",
    pin.graph.JointRevolute(np.array([0.0, 0.0, 1.0])),
    "b1",
    pin.SE3.Random(),
    "b2",
    pin.SE3.Random(),
)

g.edgeBuilder().withName("j2").withJointType(
    pin.graph.JointPrismatic(np.array([1.0, 0.0, 0.0]))
).withSourceVertex("b2").withSourcePose(pin.SE3.Random()).withTargetVertex(
    "b3"
).withTargetPose(pin.SE3.Random()).build()

# Create a Model with b1 as root body and a fixed base.
forward_model, forward_build_info = pin.graph.buildModelWithBuildInfo(g, "b1", X_I)
forward_data = forward_model.createData()

# Compute b3 placement and velocity with forward model.
qmax = np.ones(forward_model.nq)
forward_q = pin.randomConfiguration(forward_model, -qmax, qmax)
forward_v = np.random.rand(forward_model.nv)
pin.forwardKinematics(forward_model, forward_data, forward_q, forward_v)
pin.updateFramePlacements(forward_model, forward_data)

b3_index = forward_model.getFrameId("b3", pin.BODY)
X_b3 = forward_data.oMf[b3_index]
V_b3 = pin.getFrameVelocity(forward_model, forward_data, b3_index)

# Create the backward model with b3 as root body and a free flyer as root joint.
# b3 is placed at the same position and same velocity than b3 in the forward model.
# Since velocity of b3 is non zero in forward model, we have to build
# this model with a freeflyer if we want the conversion to work.
backward_model, backward_build_info = pin.graph.buildModelWithBuildInfo(
    g, "b3", X_b3, pin.graph.JointFreeFlyer()
)
backward_data = backward_model.createData()

# Create the converter and convert the configuration and velocity vector.
f_to_b_converter = pin.graph.createConverter(
    forward_model, backward_model, forward_build_info, backward_build_info
)
backward_q = f_to_b_converter.convertConfigurationVector(forward_q)
backward_v = f_to_b_converter.convertTangentVector(forward_q, forward_v)
backward_v[:6] = V_b3

pin.forwardKinematics(backward_model, backward_data, backward_q, backward_v)
pin.updateFramePlacements(backward_model, backward_data)

# Show that frame configuration and velocities are the same.
for i, frame in enumerate(forward_model.frames):
    if frame.type == pin.BODY:
        i_b = backward_model.getFrameId(frame.name, frame.type)
        motion_f = pin.getFrameVelocity(forward_model, forward_data, i)
        motion_b = pin.getFrameVelocity(backward_model, backward_data, i_b)

        print(f"Frame {frame.name} configuration and velocity in forward_model:")
        print(forward_data.oMf[i], motion_f)
        print(f"Frame {frame.name} configuration and velocity in backward_model:")
        print(backward_data.oMf[i_b], motion_b)
        print()



import pickle
import unittest
from pathlib import Path

import numpy as np
import pinocchio as pin
from test_case import PinocchioTestCase as TestCase


class TestStdMap(TestCase):
    def setUp(self):
        pass

    def test_pickle(self):
        map = pin.StdMap_String_VectorXd()
        keys = []
        for k in range(100):
            key_name = "key_" + str(k + 1)
            keys.append(key_name)
            map[key_name] = np.random.rand(10)

        pickle.dump(map, Path("save_std_map.p").open("wb"))

        map_loaded = pickle.load(Path("save_std_map.p").open("rb"))
        for key in keys:
            self.assertApprox(map[key], map_loaded[key])


if __name__ == "__main__":
    unittest.main()



"""
Example of Q-table learning with a simple discretized 1-pendulum environment using a
linear Q network.
"""

import signal
import time

import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
from dpendulum import DPendulum

# --- Random seed
RANDOM_SEED = int((time.time() % 10) * 1000)
print(f"Seed = {RANDOM_SEED}")
np.random.seed(RANDOM_SEED)
tf.set_random_seed(RANDOM_SEED)

# --- Hyper paramaters
NEPISODES = 500  # Number of training episodes
NSTEPS = 50  # Max episode length
LEARNING_RATE = 0.1  # Step length in optimizer
DECAY_RATE = 0.99  # Discount factor

# --- Environment
env = DPendulum()
NX = env.nx
NU = env.nu


# --- Q-value networks
class QValueNetwork:
    def __init__(self):
        x = tf.placeholder(shape=[1, NX], dtype=tf.float32)
        W = tf.Variable(tf.random_uniform([NX, NU], 0, 0.01, seed=100))
        qvalue = tf.matmul(x, W)
        u = tf.argmax(qvalue, 1)

        qref = tf.placeholder(shape=[1, NU], dtype=tf.float32)
        loss = tf.reduce_sum(tf.square(qref - qvalue))
        optim = tf.train.GradientDescentOptimizer(LEARNING_RATE).minimize(loss)

        self.x = x  # Network input
        self.qvalue = qvalue  # Q-value as a function of x
        self.u = u  # Policy  as a function of x
        # Reference Q-value at next step (to be set to l+Q o f)
        self.qref = qref
        self.optim = optim  # Optimizer


# --- Tensor flow initialization
tf.reset_default_graph()
qvalue = QValueNetwork()
sess = tf.InteractiveSession()
tf.global_variables_initializer().run()


def onehot(ix, n=NX):
    """Return a vector which is 0 everywhere except index <i> set to 1."""
    return np.array(
        [
            [(i == ix) for i in range(n)],
        ],
        np.float,
    )


def disturb(u, i):
    u += int(np.random.randn() * 10 / (i / 50 + 10))
    return np.clip(u, 0, NU - 1)


def rendertrial(maxiter=100):
    x = env.reset()
    for i in range(maxiter):
        u = sess.run(qvalue.u, feed_dict={qvalue.x: onehot(x)})
        x, r = env.step(u)
        env.render()
        if r == 1:
            print("Reward!")
            break


signal.signal(
    signal.SIGTSTP, lambda x, y: rendertrial()
)  # Roll-out when CTRL-Z is pressed

# --- History of search
h_rwd = []  # Learning history (for plot).

# --- Training
for episode in range(1, NEPISODES):
    x = env.reset()
    rsum = 0.0

    for step in range(NSTEPS - 1):
        # Greedy policy ...
        u = sess.run(qvalue.u, feed_dict={qvalue.x: onehot(x)})[0]
        u = disturb(u, episode)  # ... with noise
        x2, reward = env.step(u)

        # Compute reference Q-value at state x respecting HJB
        Q2 = sess.run(qvalue.qvalue, feed_dict={qvalue.x: onehot(x2)})
        Qref = sess.run(qvalue.qvalue, feed_dict={qvalue.x: onehot(x)})
        Qref[0, u] = reward + DECAY_RATE * np.max(Q2)

        # Update Q-table to better fit HJB
        sess.run(qvalue.optim, feed_dict={qvalue.x: onehot(x), qvalue.qref: Qref})

        rsum += reward
        x = x2
        if reward == 1:
            break

    h_rwd.append(rsum)
    if not episode % 20:
        print(f"Episode #{episode} done with {sum(h_rwd[-20:])} sucess")

print(f"Total rate of success: {sum(h_rwd) / NEPISODES:.3f}")
rendertrial()
plt.plot(np.cumsum(h_rwd) / range(1, NEPISODES))
plt.show()



#
# Copyright (c) 2016 CNRS
#


import lambdas
import numpy as np
import pinocchio as pin
from lambdas import (
    FCross,
    MCross,
    Mcross,
    adj,
    adjdual,
    ancestors,
    iv,
    parent,
    quad,
    td,
)
from numpy.linalg import norm
from pinocchio.robot_wrapper import RobotWrapper
from pinocchio.utils import rand


def hessian(robot, q, crossterms=False):
    """
    Compute the Hessian tensor of all the robot joints.
    If crossterms, then also compute the Si x Si terms,
    that are not part of the true Hessian but enters in some computations like VRNEA.
    """
    lambdas.setRobotArgs(robot)
    H = np.zeros([6, robot.model.nv, robot.model.nv])
    pin.computeJointJacobians(robot.model, robot.data, q)
    J = robot.data.J
    skiplast = -1 if not crossterms else None
    for joint_i in range(1, robot.model.njoints):
        for joint_j in ancestors(joint_i)[:skiplast]:  # j is a child of i
            for i in iv(joint_i):
                for j in iv(joint_j):
                    Si = J[:, i]
                    Sj = J[:, j]
                    H[:, i, j] = np.asarray(Mcross(Sj, Si))[:, 0]
    return H


# CRBA ====>       Mp[i,j] += Si.transpose()*Yk*Sj
# D-CRBA ==>   TSi Y Sj + Si Y TSj - Si YSdx Sj + Si SdxY Sj
# Term a = term d && term b = term c  --- or are null
#    -> term a is null if j<=diff
#    -> term d is null if diff>k
#    -> then a-d is nonzero only when k>=diff>j
# Due to simplification, terms cancel to the following:
#    if i<d (always true)    M[i,j,d]  =   Si.T Yd TjSd   (= -Si.T Yd Sd x Sj)
#    if j<d (possibly false) M[i,j,d] += TdSi.T Yd   Sd   (= +Si.T Sdx Yd Sj)
# where Yd is the composite inertia of the minimal subtree rooted either at j or d.


class DCRBA:
    """
    Compute the derivative (tangent application) of the mass matrix M wrt q,
    which is a NVxNVxNV tensor.
    """

    def __init__(self, robot):
        self.robot = robot
        lambdas.setRobotArgs(robot)

    def pre(self, q):
        robot = self.robot
        self.H = hessian(robot, q)
        self.J = robot.data.J.copy()
        pin.crba(robot.model, robot.data, q)
        self.Y = [
            (robot.data.oMi[i] * robot.data.Ycrb[i]).matrix()
            for i in range(0, robot.model.njoints)
        ]

        self.dM = np.zeros(
            [
                robot.model.nv,
            ]
            * 3
        )

    def __call__(self):
        robot = self.robot
        J = self.J
        Y = self.Y
        dM = self.dM
        H = self.H

        for j in range(robot.model.njoints - 1, 0, -1):
            for joint_diff in range(
                1, robot.model.njoints
            ):  # diff should be a descendant or ancestor of j
                if j not in ancestors(joint_diff) and joint_diff not in ancestors(j):
                    continue

                for i in ancestors(min(parent(joint_diff), j)):
                    i0, i1 = iv(i)[0], iv(i)[-1] + 1
                    j0, j1 = iv(j)[0], iv(j)[-1] + 1

                    Si = J[:, i0:i1]
                    Sj = J[:, j0:j1]

                    for d in iv(joint_diff):
                        T_iSd = np.array(H[:, d, i0:i1])  # this is 0 if d<=i (<=j)
                        T_jSd = np.array(H[:, d, j0:j1])  # this is 0 is d<=j

                        """
                        # d<i => TiSd=0
                        assert( norm(T_iSd)<1e-6 or not joint_diff<i )
                        # d<j => TjSd=0
                        assert( norm(T_jSd)<1e-6 or not joint_diff<j )
                        # TiSd=0 => TjSd=0
                        assert( norm(T_jSd)<1e-6 or not norm(T_iSd)<1e-6 )
                        assert( norm(T_iSd)>1e-6 )
                        """

                        Yd = Y[max(j, joint_diff)]

                        dM[i0:i1, j0:j1, d] = T_iSd.T * Yd * Sj
                        if j < joint_diff:
                            dM[i0:i1, j0:j1, d] += Si.T * Yd * T_jSd

                        # Make dM triangular by copying strict-upper triangle to lower
                        # one.
                        if i != j:
                            dM[j0:j1, i0:i1, d] = dM[i0:i1, j0:j1, d].T
                        else:
                            dM[j0:j1, i0:i1, d] += np.triu(dM[i0:i1, j0:j1, d], 1).T

        return dM


# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
class VRNEA:
    """
    Compute the C tensor so that nle(q,vq) = vq C vq.
    Noting Cv = C*vq for an arbitrary robot velocity vq:
    Since Mdot = (Cv+Cv')/2, then C = dCv/dvq and dM/dq = (Cv+Cv')

    Q is index by i,j,k i.e. tau_k = v.T*Q[:,:,k]*v
    At level k, Qk is a lower triangular matrix (same shape as H) where:
      * rows i s.t. i>k are equals to Sk.T*Ycrb_i*S_ixS_j (j the col index),
      * rows i s.t. i<=k are equals to Sk.T*Ycrb_k*S_ixS_j (j the col index)
    To avoid the call to Ycrb_i>k, the first par is set up while computing Q_k
    """

    def __init__(self, robot):
        self.robot = robot
        lambdas.setRobotArgs(robot)
        self.YJ = np.zeros([6, robot.model.nv])
        self.Q = np.zeros(
            [
                robot.model.nv,
            ]
            * 3
        )

    def __call__(self, q):
        robot = self.robot
        H = hessian(robot, q, crossterms=True)
        J = robot.data.J
        YJ = self.YJ
        Q = self.Q

        # The terms in SxS YS corresponds to f = vxYv
        # The terms in S YSxS corresponds to a = vxv  (==> f = Yvxv)
        for k in range(robot.model.njoints - 1, 0, -1):
            k0, k1 = iv(k)[0], iv(k)[-1] + 1
            Yk = (robot.data.oMi[k] * robot.data.Ycrb[k]).matrix()
            _Sk = J[:, k0:k1]
            for j in ancestors(k):  # Fill YJ = [ ... Yk*Sj ... ]_j
                j0, j1 = iv(j)[0], iv(j)[-1] + 1
                YJ[:, j0:j1] = Yk * J[:, j0:j1]

            # Fill the diagonal of the current level of Q = Q[k,:k,:k]
            for i in ancestors(k):
                i0, i1 = iv(i)[0], iv(i)[-1] + 1
                _Si = J[:, i0:i1]

                Q[k0:k1, i0:i1, i0:i1] = -td(
                    H[:, k0:k1, i0:i1], YJ[:, i0:i1], [0, 0]
                )  # = (Si x Sk)' * Yk Si

                # Fill the nondiag of the current level Q[k,:k,:k]
                for j in ancestors(i)[:-1]:
                    j0, j1 = iv(j)[0], iv(j)[-1] + 1
                    _Sj = J[:, j0:j1]

                    #  = Sk' * Yk * Si x Sj
                    Q[k0:k1, i0:i1, j0:j1] = td(
                        YJ[:, k0:k1], H[:, i0:i1, j0:j1], [0, 0]
                    )
                    #  = (Si x Sk)' * Yk Sj
                    Q[k0:k1, i0:i1, j0:j1] -= td(
                        H[:, k0:k1, i0:i1], YJ[:, j0:j1], [0, 0]
                    )
                    #  = (Sj x Sk)' * Yk Si
                    Q[k0:k1, j0:j1, i0:i1] = -td(
                        H[:, k0:k1, j0:j1], YJ[:, i0:i1], [0, 0]
                    )

            # Fill the border elements of levels below k
            # Q[kk,k,:] and Q[kk,:,k] with kk<k
            for kk in ancestors(k)[:-1]:
                kk0, kk1 = iv(kk)[0], iv(kk)[-1] + 1
                _Skk = J[:, kk0:kk1]

                for j in ancestors(k):
                    j0, j1 = iv(j)[0], iv(j)[-1] + 1
                    _Sj = J[:, j0:j1]

                    #  = Skk' Yk Sk x Sj  = (Yk Skk)' Sk x Sj
                    if k != j:
                        Q[kk0:kk1, k0:k1, j0:j1] = td(
                            YJ[:, kk0:kk1], H[:, k0:k1, j0:j1], [0, 0]
                        )
                    #  = (Sk x Skk)' Yk Sj
                    Q[kk0:kk1, k0:k1, j0:j1] += td(
                        H[:, k0:k1, kk0:kk1].T, YJ[:, j0:j1], [2, 0]
                    )
                    #  = (Sj x Skk)' Yk Sk
                    # Switch because j can be > or < than kk
                    if j < kk:
                        Q[kk0:kk1, j0:j1, k0:k1] = -Q[j0:j1, kk0:kk1, k0:k1].transpose(
                            1, 0, 2
                        )
                    elif j >= kk:
                        Q[kk0:kk1, j0:j1, k0:k1] = td(
                            H[:, j0:j1, kk0:kk1].T, YJ[:, k0:k1], [2, 0]
                        )

        return Q


# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
class Coriolis:
    def __init__(self, robot):
        self.robot = robot
        lambdas.setRobotArgs(robot)
        NV = robot.model.nv
        _NJ = robot.model.njoints
        self.C = np.zeros([NV, NV])
        self.YS = np.zeros([6, NV])
        self.Sxv = np.zeros([6, NV])

    def __call__(self, q, vq):
        robot = self.robot
        _NV = robot.model.nv
        NJ = robot.model.njoints
        C = self.C
        YS = self.YS
        Sxv = self.Sxv
        H = hessian(robot, q)

        pin.computeAllTerms(robot.model, robot.data, q, vq)
        J = robot.data.J
        oMi = robot.data.oMi
        v = [(oMi[i] * robot.data.v[i]).vector for i in range(NJ)]
        Y = [(oMi[i] * robot.data.Ycrb[i]).matrix() for i in range(NJ)]

        # --- Precomputations
        # Centroidal matrix
        for j in range(NJ):
            j0, j1 = iv(j)[0], iv(j)[-1] + 1
            YS[:, j0:j1] = Y[j] * J[:, j0:j1]

        # velocity-jacobian cross products.
        for j in range(NJ):
            j0, j1 = iv(j)[0], iv(j)[-1] + 1
            vj = v[j]
            Sxv[:, j0:j1] = adj(vj) * J[:, j0:j1]

        # --- Main loop
        for i in range(NJ - 1, 0, -1):
            i0, i1 = iv(i)[0], iv(i)[-1] + 1
            Yi = Y[i]
            Si = J[:, i0:i1]
            vp = v[robot.model.parents[i]]

            SxvY = Sxv[:, i0:i1].T * Yi
            for j in ancestors(i):
                j0, j1 = iv(j)[0], iv(j)[-1] + 1
                Sj = J[:, j0:j1]

                # COR ===> Si' Yi Sj x vj
                C[i0:i1, j0:j1] = YS[:, i0:i1].T * Sxv[:, j0:j1]
                # CEN ===> Si' vi x Yi Sj = - (Si x vi)' Yi Sj
                C[i0:i1, j0:j1] -= SxvY * Sj

            vxYS = adjdual(v[i]) * Yi * Si
            YSxv = Yi * Sxv[:, i0:i1]
            Yv = Yi * vp
            for ii in ancestors(i)[:-1]:
                ii0, ii1 = iv(ii)[0], iv(ii)[-1] + 1
                Sii = J[:, ii0:ii1]
                # COR ===> Sii' Yi Si x vi
                C[ii0:ii1, i0:i1] = Sii.T * YSxv
                # CEN ===> Sii' vi x Yi Si = (Sii x vi)' Yi Si
                C[ii0:ii1, i0:i1] += Sii.T * vxYS
                # CEN ===> Sii' Si x Yi vi = (Sii x Si)' Yi vi
                C[ii0:ii1, i0:i1] += np.array(
                    td(H[:, i0:i1, ii0:ii1], Yv, [0, 0])[:, :, 0]
                )

        return C


# --- DRNEA -------------------------------------------------------------------------
# --- DRNEA -------------------------------------------------------------------------
# --- DRNEA -------------------------------------------------------------------------


class DRNEA:
    def __init__(self, robot):
        self.robot = robot
        lambdas.setRobotArgs(robot)

    def __call__(self, q, vq, aq):
        robot = self.robot
        pin.rnea(robot.model, robot.data, q, vq, aq)
        NJ = robot.model.njoints
        NV = robot.model.nv
        J = robot.data.J
        oMi = robot.data.oMi
        v = [(oMi[i] * robot.data.v[i]).vector for i in range(NJ)]
        a = [(oMi[i] * robot.data.a_gf[i]).vector for i in range(NJ)]
        f = [(oMi[i] * robot.data.f[i]).vector for i in range(NJ)]
        Y = [(oMi[i] * robot.model.inertias[i]).matrix() for i in range(NJ)]
        Ycrb = [(oMi[i] * robot.data.Ycrb[i]).matrix() for i in range(NJ)]

        Tkf = [np.zeros([6, NV]) for i in range(NJ)]
        Yvx = [np.zeros([6, 6]) for i in range(NJ)]

        def adjf(f):
            return -np.bmat(
                [
                    [np.zeros([3, 3]), pin.skew(f[:3])],
                    [pin.skew(f[:3]), pin.skew(f[3:])],
                ]
            )

        R = self.R = np.zeros([NV, NV])

        for i in reversed(
            range(1, NJ)
        ):  # i is the torque index :    dtau_i = R[i,:] dq
            i0, i1 = iv(i)[0], iv(i)[-1] + 1
            Yi = Y[i]
            Yci = Ycrb[i]
            Si = J[:, i0:i1]
            _vi = v[i]
            _ai = a[i]
            _fi = f[i]
            _aqi = aq[i0:i1]
            vqi = vq[i0:i1]
            _dvi = Si * vqi
            li = parent(i)

            Yvx[i] += Y[i] * adj(v[i])
            Yvx[i] -= adjf(Y[i] * v[i])
            Yvx[i] -= adjdual(v[i]) * Yi

            Yvx[li] += Yvx[i]

            for k in ancestors(i):  # k is the derivative index: dtau   = R[:,k] dq_k
                k0, k1 = iv(k)[0], iv(k)[-1] + 1
                Sk = J[:, k0:k1]
                lk = parent(k)

                # Si' Yi Tk ai = Si' Yi ( Sk x (ai-alk) + (vi-vlk) x (Sk x vlk) )
                # Tk Si' Yi ai = Si' Yi ( - Sk x a[lk]  + (vi-vlk) x (Sk x vlk) )
                Tkf = Yci * (-MCross(Sk, a[lk]) + MCross(MCross(Sk, v[lk]), v[lk]))

                # Tk Si' fs = Tk Si' Ycrb[i] ai + Si' Ys (vs-vi) x (Sk x vlk)
                #           =      ""           + Si' Ys vs x (Sk x vlk) - Si' Ys vi x (Sk x vlk)  # noqa E501
                Tkf += Yvx[i] * MCross(Sk, v[lk])

                R[i0:i1, k0:k1] = Si.T * Tkf
                if i == k:
                    for kk in ancestors(k)[:-1]:
                        kk0, kk1 = iv(kk)[0], iv(kk)[-1] + 1
                        Skk = J[:, kk0:kk1]
                        R[kk0:kk1, k0:k1] = Skk.T * FCross(Sk, f[i])
                        R[kk0:kk1, k0:k1] += Skk.T * Tkf

        self.a = a
        self.v = v
        self.f = f
        self.Y = Y
        self.Ycrb = Ycrb
        self.Yvx = Yvx

        return R


# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
# --- UNIT TEST ---
if __name__ == "__main__":
    np.random.seed(0)

    robot = RobotWrapper(
        "/home/nmansard/src/pinocchio/pinocchio/models/romeo/urdf/romeo.urdf",
        [
            "/home/nmansard/src/pinocchio/pinocchio/models/romeo/",
        ],
        pin.JointModelFreeFlyer(),
    )
    q = rand(robot.model.nq)
    q[3:7] /= norm(q[3:7])
    vq = rand(robot.model.nv)

    # --- HESSIAN ---
    # --- HESSIAN ---
    # --- HESSIAN ---
    H = hessian(robot, q)

    # Compute the Hessian matrix H using RNEA, so that acor = v*H*v
    vq1 = vq * 0
    Htrue = np.zeros([6, robot.model.nv, robot.model.nv])
    for joint_i in range(1, robot.model.njoints):
        for joint_j in ancestors(joint_i)[:-1]:
            for i in iv(joint_i):
                for j in iv(joint_j):
                    vq1 *= 0
                    vq1[i] = vq1[j] = 1.0
                    pin.computeAllTerms(robot.model, robot.data, q, vq1)
                    Htrue[:, i, j] = (
                        robot.data.oMi[joint_i] * robot.data.a[joint_i]
                    ).vector.T

    print("Check hessian = \t\t", norm(H - Htrue))

    # --- dCRBA ---
    # --- dCRBA ---
    # --- dCRBA ---

    dcrba = DCRBA(robot)
    dcrba.pre(q)
    Mp = dcrba()

    # --- Validate dM/dq by finite diff
    dM = np.zeros(
        [
            robot.model.nv,
        ]
        * 3
    )
    eps = 1e-6
    dq = np.zeros(robot.model.nv)

    for diff in range(robot.model.nv):
        dM[:, :, diff] = -pin.crba(robot.model, robot.data, q)

        dq *= 0
        dq[diff] = eps
        qdq = pin.integrate(robot.model, q, dq)

        dM[:, :, diff] += pin.crba(robot.model, robot.data, qdq)

    dM /= eps

    print(
        "Check dCRBA = \t\t\t",
        max([norm(Mp[:, :, diff] - dM[:, :, diff]) for diff in range(robot.model.nv)]),
    )

    # --- vRNEA ---
    # --- vRNEA ---
    # --- vRNEA ---

    vrnea = VRNEA(robot)
    Q = vrnea(q)

    # --- Compute C from rnea, for future check
    robot.model.gravity = pin.Motion.Zero()

    def rnea0(q, vq):
        return pin.nle(robot.model, robot.data, q, vq)

    vq1 = vq * 0
    C = np.zeros(
        [
            robot.model.nv,
        ]
        * 3
    )
    for i in range(robot.model.nv):
        vq1 *= 0
        vq1[i] = 1
        C[:, i, i] = rnea0(q, vq1).T

    for i in range(robot.model.nv):
        for j in range(robot.model.nv):
            if i == j:
                continue
            vq1 *= 0
            vq1[i] = vq1[j] = 1.0
            C[:, i, j] = (rnea0(q, vq1).T - C[:, i, i] - C[:, j, j]) / 2

    print("Check d/dv rnea = \t\t", norm(quad(Q, vq) - rnea0(q, vq)))
    print("Check C  = Q+Q.T = \t\t", norm((Q + Q.transpose(0, 2, 1)) / 2 - C))
    print("Check dM = C+C.T /2 \t\t", norm(Mp - (C + C.transpose(1, 0, 2))))
    print(
        "Check dM = Q+Q.T+Q.T+Q.T /2 \t",
        norm(
            Mp
            - (Q + Q.transpose(0, 2, 1) + Q.transpose(1, 0, 2) + Q.transpose(2, 0, 1))
            / 2
        ),
    )

    # --- CORIOLIS
    # --- CORIOLIS
    # --- CORIOLIS
    coriolis = Coriolis(robot)
    C = coriolis(q, vq)
    print("Check coriolis \t\t\t", norm(C * vq - rnea0(q, vq)))

    # --- DRNEA
    # --- DRNEA
    # --- DRNEA
    drnea = DRNEA(robot)
    aq = rand(robot.model.nv)
    R = drnea(q, vq, aq)

    NV = robot.model.nv
    Rd = np.zeros([NV, NV])
    eps = 1e-8
    r0 = pin.rnea(robot.model, robot.data, q, vq, aq).copy()
    for i in range(NV):
        dq = np.zeros(NV)
        dq[i] = eps
        qdq = pin.integrate(robot.model, q, dq)
        Rd[:, i] = (pin.rnea(robot.model, robot.data, qdq, vq, aq) - r0) / eps
    print("Check drnea    \t\t\t", norm(Rd - R))



import numpy as np
import pinocchio as pin

##
## In this short script, we show how to compute the derivatives of the
## inverse dynamics (RNEA), using the algorithms proposed in:
##
## Analytical Derivatives of Rigid Body Dynamics Algorithms,
## Justin Carpentier and Nicolas Mansard,
## Robotics: Science and Systems, 2018
##

# Create model and data

model = pin.buildSampleModelHumanoidRandom()
data = model.createData()

# Set bounds (by default they are undefinded for a the Simple Humanoid model)

model.lowerPositionLimit = -np.ones((model.nq, 1))
model.upperPositionLimit = np.ones((model.nq, 1))

q = pin.randomConfiguration(model)  # joint configuration
v = np.random.rand(model.nv, 1)  # joint velocity
a = np.random.rand(model.nv, 1)  # joint acceleration

# Evaluate the derivatives

pin.computeRNEADerivatives(model, data, q, v, a)

# Retrieve the derivatives in data

dtau_dq = data.dtau_dq  # Derivatives of the ID w.r.t. the joint config vector
dtau_dv = data.dtau_dv  # Derivatives of the ID w.r.t. the joint velocity vector
dtau_da = data.M  # Derivatives of the ID w.r.t. the joint acceleration vector



import os
import unittest
from pathlib import Path

import pinocchio as pin


def checkGeom(geom1, geom2):
    return geom1.ngeoms == geom2.ngeoms


@unittest.skipUnless(pin.WITH_URDFDOM, "Needs URDFDOM")
class TestGeometryObjectUrdfBindings(unittest.TestCase):
    def setUp(self):
        self.model_dir = Path(os.environ.get("EXAMPLE_ROBOT_DATA_MODEL_DIR"))
        self.mesh_path = self.model_dir.parent.parent
        self.model_path = self.model_dir / "romeo_description/urdf/romeo.urdf"

    def test_load(self):
        hint_list = [self.mesh_path, "wrong/hint"]
        expected_mesh_path = (
            self.model_dir / "romeo_description/meshes/V1/collision/LHipPitch.dae"
        )

        model = pin.buildModelFromUrdf(self.model_path, pin.JointModelFreeFlyer())
        collision_model = pin.buildGeomFromUrdf(
            model, self.model_path, pin.GeometryType.COLLISION, hint_list
        )

        col = collision_model.geometryObjects[1]
        self.assertEqual(
            os.path.normpath(col.meshPath), os.path.normpath(expected_mesh_path)
        )

    def test_self_load(self):
        hint_list = [self.mesh_path]

        model = pin.buildModelFromUrdf(self.model_path, pin.JointModelFreeFlyer())
        collision_model_ref = pin.buildGeomFromUrdf(
            model, self.model_path, pin.GeometryType.COLLISION, hint_list
        )

        collision_model_self = pin.GeometryModel()
        pin.buildGeomFromUrdf(
            model,
            self.model_path,
            pin.GeometryType.COLLISION,
            collision_model_self,
            hint_list,
        )
        self.assertTrue(checkGeom(collision_model_ref, collision_model_self))

        collision_model_self = pin.GeometryModel()
        pin.buildGeomFromUrdf(
            model,
            self.model_path,
            pin.GeometryType.COLLISION,
            collision_model_self,
            self.mesh_path,
        )
        self.assertTrue(checkGeom(collision_model_ref, collision_model_self))

        hint_vec = [self.mesh_path]

        collision_model_self = pin.GeometryModel()
        pin.buildGeomFromUrdf(
            model,
            self.model_path,
            pin.GeometryType.COLLISION,
            collision_model_self,
            hint_vec,
        )
        self.assertTrue(checkGeom(collision_model_ref, collision_model_self))

    def test_multi_load(self):
        hint_list = [self.mesh_path, "wrong/hint"]
        expected_collision_path = (
            self.model_dir / "romeo_description/meshes/V1/collision/LHipPitch.dae"
        )
        expected_visual_path = (
            self.model_dir / "romeo_description/meshes/V1/visual/LHipPitch.dae"
        )

        model = pin.buildModelFromUrdf(self.model_path, pin.JointModelFreeFlyer())

        collision_model = pin.buildGeomFromUrdf(
            model, self.model_path, pin.GeometryType.COLLISION, hint_list
        )
        col = collision_model.geometryObjects[1]
        self.assertEqual(
            os.path.normpath(col.meshPath), os.path.normpath(expected_collision_path)
        )

        visual_model = pin.buildGeomFromUrdf(
            model, self.model_path, pin.GeometryType.VISUAL, hint_list
        )
        vis = visual_model.geometryObjects[1]
        self.assertEqual(
            os.path.normpath(vis.meshPath), os.path.normpath(expected_visual_path)
        )

        model_2, collision_model_2, visual_model_2 = pin.buildModelsFromUrdf(
            self.model_path, hint_list, pin.JointModelFreeFlyer()
        )

        self.assertEqual(model, model_2)

        col_2 = collision_model_2.geometryObjects[1]
        self.assertEqual(
            os.path.normpath(col_2.meshPath), os.path.normpath(expected_collision_path)
        )

        vis_2 = visual_model_2.geometryObjects[1]
        self.assertEqual(
            os.path.normpath(vis_2.meshPath), os.path.normpath(expected_visual_path)
        )

        model_c, collision_model_c = pin.buildModelsFromUrdf(
            self.model_path,
            hint_list,
            pin.JointModelFreeFlyer(),
            geometry_types=pin.GeometryType.COLLISION,
        )

        self.assertEqual(model, model_c)

        col_c = collision_model_c.geometryObjects[1]
        self.assertEqual(
            os.path.normpath(col_c.meshPath), os.path.normpath(expected_collision_path)
        )

        model_v, visual_model_v = pin.buildModelsFromUrdf(
            self.model_path,
            hint_list,
            pin.JointModelFreeFlyer(),
            geometry_types=pin.GeometryType.VISUAL,
        )

        self.assertEqual(model, model_v)

        vis_v = visual_model_v.geometryObjects[1]
        self.assertEqual(
            os.path.normpath(vis_v.meshPath), os.path.normpath(expected_visual_path)
        )


if __name__ == "__main__":
    unittest.main()



import unittest

import pinocchio as pin
from test_case import PinocchioTestCase as TestCase


class TestKinematicRegressorBindings(TestCase):
    def test_all(self):
        model = pin.buildSampleModelHumanoidRandom()

        joint_name = "larm6_joint"
        joint_id = model.getJointId(joint_name)
        frame_id = model.addBodyFrame("test_body", joint_id, pin.SE3.Identity(), -1)

        data = model.createData()

        model.lowerPositionLimit[:7] = -1.0
        model.upperPositionLimit[:7] = 1.0

        q = pin.randomConfiguration(model)
        pin.forwardKinematics(model, data, q)

        R1 = pin.computeJointKinematicRegressor(
            model, data, joint_id, pin.ReferenceFrame.LOCAL, pin.SE3.Identity()
        )
        R2 = pin.computeJointKinematicRegressor(
            model, data, joint_id, pin.ReferenceFrame.LOCAL
        )

        self.assertApprox(R1, R2)

        R3 = pin.computeFrameKinematicRegressor(
            model, data, frame_id, pin.ReferenceFrame.LOCAL
        )
        self.assertApprox(R1, R3)


if __name__ == "__main__":
    unittest.main()



import unittest

import pinocchio as pin
from pinocchio.utils import np, npl, rand, zero
from test_case import PinocchioTestCase as TestCase

# This whole file seems to be outdated and superseded by more recent tests
# Probably it should be removed and its contents moved or split somewhere else


class TestSE3(TestCase):
    def setUp(self):
        self.R = rand([3, 3])
        self.R, _, _ = npl.svd(self.R)
        self.p = rand(3)
        self.m = pin.SE3(self.R, self.p)

    def test_se3(self):
        R, p, m = self.R, self.p, self.m
        X = np.vstack(
            [np.hstack([R, pin.skew(p).dot(R)]), np.hstack([zero([3, 3]), R])]
        )
        self.assertApprox(m.action, X)
        M = np.vstack(
            [np.hstack([R, np.expand_dims(p, 1)]), np.array([[0.0, 0.0, 0.0, 1.0]])]
        )
        self.assertApprox(m.homogeneous, M)
        m2 = pin.SE3.Random()
        self.assertApprox((m * m2).homogeneous, m.homogeneous.dot(m2.homogeneous))
        self.assertApprox((~m).homogeneous, npl.inv(m.homogeneous))

        p = rand(3)
        self.assertApprox(m * p, m.rotation.dot(p) + m.translation)
        self.assertApprox(
            m.actInv(p), m.rotation.T.dot(p) - m.rotation.T.dot(m.translation)
        )

        # Currently, the different cases do not throw the same exception type.
        # To have a more robust test, only Exception is checked.
        # In the comments, the most specific actual exception class at the time of
        # writing
        p = rand(5)
        with self.assertRaises(Exception):  # RuntimeError
            m * p
        with self.assertRaises(Exception):  # RuntimeError
            m.actInv(p)
        with self.assertRaises(
            Exception
        ):  # Boost.Python.ArgumentError (subclass of TypeError)
            m.actInv("42")

    def test_motion(self):
        m = self.m
        self.assertApprox(pin.Motion.Zero().vector, zero(6))
        v = pin.Motion.Random()
        self.assertApprox((m * v).vector, m.action.dot(v.vector))
        self.assertApprox((m.actInv(v)).vector, npl.inv(m.action).dot(v.vector))
        vv = v.linear
        vw = v.angular
        self.assertApprox(v.vector, np.concatenate([vv, vw]))
        self.assertApprox((v ^ v).vector, zero(6))

    def test_force(self):
        m = self.m
        self.assertApprox(pin.Force.Zero().vector, zero(6))
        f = pin.Force.Random()
        ff = f.linear
        ft = f.angular
        self.assertApprox(f.vector, np.concatenate([ff, ft]))

        self.assertApprox((m * f).vector, npl.inv(m.action.T).dot(f.vector))
        self.assertApprox((m.actInv(f)).vector, m.action.T.dot(f.vector))
        v = pin.Motion.Random()
        f = pin.Force(np.concatenate([v.vector[3:], v.vector[:3]]))
        self.assertApprox((v ^ f).vector, zero(6))

    def test_inertia(self):
        m = self.m
        Y1 = pin.Inertia.Random()
        Y2 = pin.Inertia.Random()
        Y = Y1 + Y2
        self.assertApprox(Y1.matrix() + Y2.matrix(), Y.matrix())
        v = pin.Motion.Random()
        self.assertApprox((Y * v).vector, Y.matrix().dot(v.vector))
        self.assertApprox(
            (m * Y).matrix(),
            m.inverse().action.T.dot(Y.matrix()).dot(m.inverse().action),
        )
        self.assertApprox(
            (m.actInv(Y)).matrix(), m.action.T.dot(Y.matrix()).dot(m.action)
        )

    def test_cross(self):
        m = pin.Motion.Random()
        f = pin.Force.Random()
        self.assertApprox(m ^ m, m.cross(m))
        self.assertApprox(m ^ f, m.cross(f))
        with self.assertRaises(TypeError):
            m ^ 2


if __name__ == "__main__":
    unittest.main()



import time

from pinocchio.utils import se3ToXYZQUATtuple


def display_prm(robot, graph):
    """
    Take a graph object containing a list of configurations q and a dictionnary of graph
    relations edge. Display the configurations by the correspond placement of the robot
    end effector. Display the graph relation by vertices connecting the robot end
    effector positions.
    """

    gui = robot.viewer.gui

    try:
        gui.deleteNode("world/prm", True)
    except:  # noqa: E722
        pass
    gui.createRoadmap(
        "world/prm", [1.0, 0.2, 0.2, 0.8], 1e-2, 1e-2, [1.0, 0.2, 0.2, 0.8]
    )

    for q in graph.q:
        gui.addNodeToRoadmap("world/prm", se3ToXYZQUATtuple(robot.position(q, 6)))

        for parent, children in graph.children.items():
            for child in children:
                if child > parent:
                    q1, q2 = graph.q[parent], graph.q[child]
                    p1 = robot.position(q1, 6).translation.ravel().tolist()[0]
                    p2 = robot.position(q2, 6).translation.ravel().tolist()[0]
                    gui.addEdgeToRoadmap("world/prm", p1, p2)

                    gui.refresh()


def display_path(robot, path, sleeptime=1e-2):
    """
    Display a path, i.e. a sequence of robot configuration, by moving the robot
    to each list element.
    """
    for q in path:
        robot.display(q)
        time.sleep(sleeptime)



#!/usr/bin/env python
#
# Copyright (C) 2023 Inria

import numpy as np

try:
    from robot_descriptions.loaders.pinocchio import load_robot_description
except ModuleNotFoundError:
    print("This example loads robot descriptions from robot_descriptions.py:")
    print("\n\tpip install robot_descriptions")

print("Goal: load a legged robot from its URDF, then modify leg lengths")

print("Loading robot description from URDF...")
robot = load_robot_description("upkie_description")
model = robot.model

# Joint placement offsets compared to the leg length
# We know from the robot description that both are along the local y-axis
known_offsets = {"knee": 0.072, "wheel": 0.065}


def check_limb_lengths(limb_length: float) -> bool:
    print(f"Checking that limbs are {limb_length} m long... ", end="")
    for side in ("left", "right"):
        for joint in ("knee", "wheel"):
            joint_id = model.getJointId(f"{side}_{joint}")
            if not np.allclose(
                model.jointPlacements[joint_id].translation[1],
                limb_length - known_offsets[joint],
            ):
                print("{side}_{joint} placement is wrong!")
                return False
    return True


# We know from the description that the leg length should be 24 cm
if check_limb_lengths(0.24):
    print("OK, the model is as we expect it")


def update_limb_lengths(length: float) -> None:
    """Update femur and tibia lengths in the robot model.

    Args:
        length: New femur and tibia length, in meters.
    """
    for side in ("left", "right"):
        for joint in ("knee", "wheel"):
            joint_id = model.getJointId(f"{side}_{joint}")
            model.jointPlacements[joint_id].translation[1] = (
                length - known_offsets[joint]
            )


update_limb_lengths(0.3)  # update femurs and tibias to 30 cm
if check_limb_lengths(0.3):
    print("OK, the update worked!")



import unittest

import numpy as np
import pinocchio as pin
from pinocchio.utils import rand


def df_dq(model, func, q, h=1e-9):
    """Perform df/dq by num_diff. q is in the lie manifold.
    :params func: function to differentiate f : np.array -> np.array
    :params q: configuration value at which f is differentiated. type np.array
    :params h: eps

    :returns df/dq
    """
    dq = np.zeros(model.nv)
    f0 = func(q)
    res = np.zeros([len(f0), model.nv])
    for iq in range(model.nv):
        dq[iq] = h
        res[:, iq] = (func(pin.integrate(model, q, dq)) - f0) / h
        dq[iq] = 0
    return res


class TestVComDerivativesBindings(unittest.TestCase):
    def setUp(self):
        self.rmodel = rmodel = pin.buildSampleModelHumanoid()
        self.rdata = rmodel.createData()
        self.rdata_fd = rmodel.createData()

        self.rmodel.lowerPositionLimit[:3] = -1.0
        self.rmodel.upperPositionLimit[:3] = -1.0
        self.q = pin.randomConfiguration(rmodel)
        self.vq = rand(rmodel.nv) * 2 - 1

        self.precision = 1e-8

    def test_numdiff(self):
        rmodel, rdata = self.rmodel, self.rdata
        rdata_fd = self.rdata_fd
        q, vq = self.q, self.vq

        #### Compute d/dq VCOM with the algo.
        pin.computeAllTerms(rmodel, rdata, q, vq)
        dvc_dq = pin.getCenterOfMassVelocityDerivatives(rmodel, rdata)

        #### Approximate d/dq VCOM by finite diff.
        def calc_vc(q, vq):
            """Compute COM velocity"""
            pin.centerOfMass(rmodel, rdata_fd, q, vq)
            return rdata_fd.vcom[0].copy()

        dvc_dqn = df_dq(rmodel, lambda _q: calc_vc(_q, vq), q)

        self.assertTrue(np.allclose(dvc_dq, dvc_dqn, atol=np.sqrt(self.precision)))


if __name__ == "__main__":
    unittest.main()



class Graph:
    def __init__(self):
        self.children = {}  # dictionnary giving the list of childrens for each node.
        self.q = []  # configuration associated to each node.
        self.connex = []  # ID of the connex component the node is belonging to.
        self.nconnex = 0  # number of connex components.
        self.existing_connex = []  # List of existing connex component ID.

    def add_node(self, q=None, new_connex=False):
        """
        Create the memory to store a new edge. Initialize all components to None.
        Create an empty list of children.
        """
        idx = len(self.children)
        self.children[idx] = []
        self.q.append(q)
        self.connex.append(None)
        if new_connex:
            self.new_connex(idx)
        return idx

    def add_edge(self, first, second, orientation=0):
        """
        Add edge from first to second. Also add edge from second to first if orientation
        is null.
        """
        assert first in self.children and second in self.children
        if orientation >= 0:
            self.children[first].append(second)
        if orientation <= 0:
            self.children[second].append(first)

    def new_connex(self, idx):
        """
        Create a new connex component for node <idx>
        """
        self.connex[idx] = self.nconnex
        self.existing_connex.append(self.nconnex)
        self.nconnex += 1

    def rename_connex(self, past, future):
        """
        Change the index of the all the nodes belonging to a connex component.
        Useful when merging two connex components.
        """
        try:
            self.existing_connex.remove(past)
            self.connex = [c if c != past else future for c in self.connex]
        except:  # noqa: E722
            pass

    def connexIndexes(self, connex):
        """
        Return the list of all node indexes belonging to connex component <connex>.
        """
        return [i for i, c in enumerate(self.connex) if c == connex]



import os
from pathlib import Path

import numpy as np
import pinocchio as pin

# Goal: Build a reduced model from an existing URDF model by fixing the desired joints
# at a specified position.

# Load UR robot arm
# This path refers to Pinocchio source code but you can define your own directory here.
model_path = Path(os.environ.get("EXAMPLE_ROBOT_DATA_MODEL_DIR"))
mesh_dir = model_path.parent.parent
# You should change here to set up your own URDF file
urdf_filename = model_path / "ur_description/urdf/ur5_robot.urdf"
model, collision_model, visual_model = pin.buildModelsFromUrdf(urdf_filename, mesh_dir)

# Check dimensions of the original model
print("standard model: dim=" + str(len(model.joints)))
for jn in model.joints:
    print(jn)
print("-" * 30)

# Create a list of joints to lock
jointsToLock = ["wrist_1_joint", "wrist_2_joint", "wrist_3_joint"]

# Get the ID of all existing joints
jointsToLockIDs = []
for jn in jointsToLock:
    if model.existJointName(jn):
        jointsToLockIDs.append(model.getJointId(jn))
    else:
        print("Warning: joint " + str(jn) + " does not belong to the model!")

# Set initial position of both fixed and revoulte joints
initialJointConfig = np.array(
    [
        0,
        0,
        0,  # shoulder and elbow
        1,
        1,
        1,
    ]
)  # gripper)

# Option 1: Only build the reduced model in case no display needed:
model_reduced = pin.buildReducedModel(model, jointsToLockIDs, initialJointConfig)

# Option 2: Build the reduced model including the geometric model for proper displaying
# of the robot.
model_reduced, visual_model_reduced = pin.buildReducedModel(
    model, visual_model, jointsToLockIDs, initialJointConfig
)

# Option 3: Build the reduced model including multiple geometric models (for example:
# visuals, collision).
geom_models = [visual_model, collision_model]
model_reduced, geometric_models_reduced = pin.buildReducedModel(
    model,
    list_of_geom_models=geom_models,
    list_of_joints_to_lock=jointsToLockIDs,
    reference_configuration=initialJointConfig,
)
# geometric_models_reduced is a list, ordered as the passed variable "geom_models" so:
visual_model_reduced, collision_model_reduced = (
    geometric_models_reduced[0],
    geometric_models_reduced[1],
)

# Check dimensions of the reduced model
# options 1-3 only take joint ids
print("joints to lock (only ids):", jointsToLockIDs)
print("reduced model: dim=" + str(len(model_reduced.joints)))
print("-" * 30)

# Option 4: Build a reduced model of a robot using RobotWrapper
# reference_configuration is optional: if not provided, neutral configuration used
# you can even mix joint names and joint ids
mixed_jointsToLockIDs = [jointsToLockIDs[0], "wrist_2_joint", "wrist_3_joint"]
robot = pin.RobotWrapper.BuildFromURDF(urdf_filename, mesh_dir)
reduced_robot = robot.buildReducedRobot(
    list_of_joints_to_lock=mixed_jointsToLockIDs,
    reference_configuration=initialJointConfig,
)

# Check dimensions of the reduced model and joint info
print("mixed joints to lock (names and ids):", mixed_jointsToLockIDs)
print("RobotWrapper reduced model: dim=" + str(len(reduced_robot.model.joints)))
for jn in robot.model.joints:
    print(jn)



from time import sleep

import hppfcl
import numpy as np
import pinocchio as pin
from pinocchio.visualize import MeshcatVisualizer


def create_pin_cube_model(j0="freeflyer"):
    model = pin.Model()

    if j0 == "freeflyer":
        j0 = pin.JointModelFreeFlyer()
    elif j0 == "spherical":
        j0 = pin.JointModelSpherical()
    elif j0 == "sphericalzyx":
        j0 = pin.JointModelSphericalZYX()
    else:
        raise ValueError("Unknown joint type")

    jointCube = model.addJoint(0, j0, pin.SE3.Identity(), "joint0")
    M = pin.SE3.Identity()
    model.appendBodyToJoint(jointCube, pin.Inertia.FromBox(1, 0.8, 0.4, 0.2), M)
    return model


def create_pin_geometry_cube_model(model):
    jointCube = model.getFrameId("joint0")
    geom_model = pin.GeometryModel()
    cube_shape = hppfcl.Box(0.8, 0.4, 0.2)  # x, y, z
    cube = pin.GeometryObject(
        "cube_shape", 0, jointCube, cube_shape, pin.SE3.Identity()
    )
    cube.meshColor = np.array([1.0, 0.1, 0.1, 0.5])
    geom_model.addGeometryObject(cube)
    return geom_model


def create_model(joint0_type="freeflyer"):
    model = create_pin_cube_model(joint0_type)
    geom_model = create_pin_geometry_cube_model(model)
    return model, geom_model


def pin_step(model, vizer, v_index_increment, dt=0.1):
    q = pin.neutral(model)
    v = np.zeros(model.nv)
    v[v_index_increment] += 1.0

    vizer.display(q)
    print(f"{q=}, {v=}")

    sleep(1)

    q_next = pin.integrate(model, q, v * dt)

    print(f"{q_next=}")
    return q_next


if __name__ == "__main__":
    list_joint0 = ["freeflyer", "sphericalzyx", "spherical"]

    for joint0_name in list_joint0:
        print(f"\njoint0_name = {joint0_name}")

        model, geom_model = create_model(joint0_name)
        print(f"{model.nq=}")
        print(f"{model.nv=}")

        q = pin.neutral(model)
        print(f"neutral q configuration: {q=}")

        vizer = MeshcatVisualizer(model, geom_model, geom_model)
        vizer.initViewer(open=True, loadModel=True)
        vizer.display(q)

        sleep(2)

        for i in range(model.nv):
            print(f"v[{i}] += 1")
            q_next = pin_step(model, vizer, i, dt=0.2)
            vizer.display(q_next)
            sleep(2)



# This examples shows how to load and move a robot in Viser.
# Note: this feature requires Viser to be installed, this can be done using
# pip install --user viser

import os
import sys
from pathlib import Path

import numpy as np
import pinocchio as pin
from pinocchio.visualize import ViserVisualizer

# Load the URDF model.
# Conversion with str seems to be necessary when executing this file with ipython

model_path = Path(os.environ.get("EXAMPLE_ROBOT_DATA_MODEL_DIR"))
mesh_dir = model_path.parent.parent
# urdf_filename = "talos_reduced.urdf"
# urdf_model_path = join(join(model_path,"talos_data/robots"),urdf_filename)
urdf_filename = "solo.urdf"
urdf_model_path = model_path / "solo_description/robots" / urdf_filename

model, collision_model, visual_model = pin.buildModelsFromUrdf(
    urdf_model_path, mesh_dir, pin.JointModelFreeFlyer()
)

# Start a new Viser server and client.
try:
    viz = ViserVisualizer(model, collision_model, visual_model)
    viz.initViewer(open=True)
except ImportError as err:
    print("Error while initializing the viewer. It seems you should install viser")
    print(err)
    sys.exit(0)

# Create a convex shape from solo main body
mesh = visual_model.geometryObjects[0].geometry
mesh.buildConvexRepresentation(True)
convex = mesh.convex

# Place the convex object on the scene and display it
if convex is not None:
    placement = pin.SE3.Identity()
    placement.translation[0] = 2.0
    geometry = pin.GeometryObject("convex", 0, placement, convex)
    geometry.meshColor = np.ones(4)
    visual_model.addGeometryObject(geometry)
    # After modifying the visual_model we must rebuild
    # associated data inside the visualizer
    viz.rebuildData()

# Load the robot in the viewer.
viz.loadViewerModel()

# Display a robot configuration.
q0 = pin.neutral(model)
viz.display(q0)
viz.displayVisuals(True)

# Display another robot.
viz2 = ViserVisualizer(model, collision_model, visual_model)
viz2.initViewer(viz.viewer)
viz2.loadViewerModel(rootNodeName="pinocchio2")
q = q0.copy()
q[1] = 1.0
viz2.display(q)

# standing config
q1 = np.array(
    [0.0, 0.0, 0.235, 0.0, 0.0, 0.0, 1.0, 0.8, -1.6, 0.8, -1.6, -0.8, 1.6, -0.8, 1.6]
)

v0 = np.random.randn(model.nv) * 2
data = viz.data
pin.forwardKinematics(model, data, q1, v0)
frame_id = model.getFrameId("HR_FOOT")
viz.display()

model.gravity.linear[:] = 0.0
dt = 0.01


def sim_loop():
    tau0 = np.zeros(model.nv)
    qs = [q1]
    vs = [v0]
    nsteps = 100
    for i in range(nsteps):
        q = qs[i]
        v = vs[i]
        a1 = pin.aba(model, data, q, v, tau0)
        vnext = v + dt * a1
        qnext = pin.integrate(model, q, dt * vnext)
        qs.append(qnext)
        vs.append(vnext)
        viz.display(qnext)
    return qs, vs


qs, vs = sim_loop()

with viz.create_video_ctx("../leap.mp4"):
    viz.play(qs, dt)



#
# Copyright (c) 2016-2023 CNRS INRIA
#

from math import pi
from typing import ClassVar

import pinocchio as pin
from pinocchio.utils import np, rotate

DENSITY = 1


def placement(x=0, y=0, z=0, rx=0, ry=0, rz=0):
    m = pin.SE3.Identity()
    m.translation = np.array([x, y, z])
    m.rotation *= rotate("x", rx)
    m.rotation *= rotate("y", ry)
    m.rotation *= rotate("z", rz)
    return m


def color(body_number=1):
    return [int(i) for i in f"{int(bin(body_number % 8)[2:]):03d}"] + [1]


class ModelWrapper:
    def __init__(self, name=None, display=False):
        self.visuals = [
            ("universe", pin.SE3.Identity(), pin.SE3.Identity().translation)
        ]
        self.name = self.__class__.__name__ if name is None else name
        self.model = pin.Model()
        self.display = display
        self.add_joints()

    def add_joints(self):
        for joint in self.joints:
            self.add_joint(**joint)

    def add_joint(
        self,
        joint_name,
        joint_model=None,
        joint_placement=None,
        lever=None,
        shape="box",
        dimensions=1,
        mass=None,
        body_color=1,
        parent=0,
    ):
        if joint_model is None:
            joint_model = pin.JointModelFreeFlyer()
        elif isinstance(joint_model, str):
            joint_model = pin.__dict__["JointModel" + joint_model]()
        if joint_placement is None:
            joint_placement = pin.SE3.Identity()
        elif isinstance(joint_placement, dict):
            joint_placement = placement(**joint_placement)
        if lever is None:
            lever = pin.SE3.Identity()
        elif isinstance(lever, dict):
            lever = placement(**lever)
        joint_name, body_name = (
            f"world/{self.name}_{joint_name}_{i}" for i in ("joint", "body")
        )
        body_inertia = pin.Inertia.Random()
        if shape == "box":
            w, h, d = (
                (float(i) for i in dimensions)
                if isinstance(dimensions, tuple)
                else [float(dimensions)] * 3
            )
            if mass is None:
                mass = w * h * d * DENSITY
            body_inertia = pin.Inertia.FromBox(mass, w, h, d)
            if self.display:
                self.display.viewer.gui.addBox(body_name, w, h, d, color(body_color))
        elif shape == "cylinder":
            r, h = dimensions
            if mass is None:
                mass = pi * r**2 * h * DENSITY
            body_inertia = pin.Inertia.FromCylinder(mass, r, h)
            if self.display:
                self.display.viewer.gui.addCylinder(body_name, r, h, color(body_color))
        elif shape == "sphere":
            w, h, d = (
                (float(i) for i in dimensions)
                if isinstance(dimensions, tuple)
                else [float(dimensions)] * 3
            )
            if mass is None:
                mass = 4.0 / 3.0 * pi * w * h * d * DENSITY
            body_inertia = pin.Inertia.FromEllipsoid(mass, w, h, d)
            if self.display:
                self.display.viewer.gui.addSphere(
                    body_name, dimensions, color(body_color)
                )
        body_inertia.lever = lever.translation
        joint_id = self.model.addJoint(parent, joint_model, joint_placement, joint_name)
        self.model.appendBodyToJoint(joint_id, body_inertia, pin.SE3.Identity())
        self.model.addJointFrame(joint_id, -1)
        self.model.addBodyFrame(body_name, joint_id, pin.SE3.Identity(), -1)
        self.visuals.append((body_name, joint_placement, lever))
        self.data = self.model.createData()
        if self.display:
            self.place()

    def place(self):
        for i, (name, placement, lever) in enumerate(self.visuals):
            if i == 0:
                continue
            self.display.place(name, self.data.oMi[i] * placement * lever)
        self.display.viewer.gui.refresh()


class RobotDisplay:
    def __init__(self, window_name="pinocchio"):
        from gepetto import corbaserver

        self.viewer = corbaserver.Client()
        try:
            window_id = self.viewer.gui.getWindowID(window_name)
        except:  # noqa: E722
            window_id = self.viewer.gui.createWindow(window_name)
            self.viewer.gui.createSceneWithFloor("world")
            self.viewer.gui.addSceneToWindow("world", window_id)
        self.viewer.gui.setLightingMode("world", "OFF")
        self.viewer.gui.setVisibility("world/floor", "OFF")
        self.viewer.gui.refresh()

    def place(self, obj_name, m):
        self.viewer.gui.applyConfiguration(obj_name, pin.se3ToXYZQUATtuple(m))


class SimplestWalker(ModelWrapper):
    joints: ClassVar = [
        {
            "joint_name": "pelvis",
            "dimensions": (0.1, 0.2, 0.1),
            "mass": 0.5,
        },
        {
            "joint_name": "left_leg",
            "joint_model": "RY",
            "joint_placement": {"y": -0.15},
            "lever": {"z": -0.45},
            "shape": "cylinder",
            "dimensions": (0.1, 0.9),
            "mass": 20,
            "body_color": 2,
            "parent": 1,
        },
        {
            "joint_name": "right_leg",
            "joint_model": "RY",
            "joint_placement": {"y": 0.15},
            "lever": {"z": -0.45},
            "shape": "cylinder",
            "dimensions": (0.1, 0.9),
            "mass": 20,
            "body_color": 3,
            "parent": 1,
        },
    ]


walker = SimplestWalker()
model = walker.model



import unittest

import pinocchio as pin


class TestGeometryModelBindings(unittest.TestCase):
    def test_pair_equals(self):
        c1 = pin.CollisionPair(1, 2)
        c2 = pin.CollisionPair(1, 2)
        c3 = pin.CollisionPair(3, 4)

        self.assertEqual(c1, c2)
        self.assertTrue(c1 == c2)
        self.assertFalse(c1 != c2)

        self.assertNotEqual(c1, c3)
        self.assertTrue(c1 != c3)
        self.assertFalse(c1 == c3)

    def test_pair_copy(self):
        c1 = pin.CollisionPair(1, 2)
        c2 = c1.copy()

        self.assertEqual(c1, c2)

        c2.second = 3
        self.assertNotEqual(c1, c2)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020 INRIA
#
# ruff: noqa: F401, F403, F405

# Manually register submodules
import sys

from .. import utils
from ..explog import exp, log
from ..pinocchio_pywrap_casadi import *
from ..pinocchio_pywrap_casadi import __raw_version__, __version__

sys.modules["pinocchio.casadi.rpy"] = rpy
sys.modules["pinocchio.casadi.cholesky"] = cholesky

if WITH_HPP_FCL:
    try:
        import hppfcl
        from hppfcl import (
            CachedMeshLoader,
            CollisionGeometry,
            CollisionResult,
            Contact,
            DistanceResult,
            MeshLoader,
            StdVec_CollisionResult,
            StdVec_Contact,
            StdVec_DistanceResult,
        )

        WITH_HPP_FCL_BINDINGS = True
    except ImportError:
        WITH_HPP_FCL_BINDINGS = False
else:
    WITH_HPP_FCL_BINDINGS = False



# ruff: noqa: E501
#
# Copyright (c) 2018-2020 CNRS INRIA
#

## In this file, some shortcuts are provided ##

from . import WITH_HPP_FCL, WITH_HPP_FCL_BINDINGS
from . import pinocchio_pywrap_default as pin

nle = pin.nonLinearEffects


def buildModelsFromUrdf(
    filename, *args, **kwargs
) -> tuple[pin.Model, pin.GeometryModel, pin.GeometryModel]:
    """Parse the URDF file given in input and return a Pinocchio Model followed by corresponding GeometryModels of types specified by geometry_types, in the same order as listed.
    Arguments:
        - filename - name of the urdf file to load
        - package_dirs - where the meshes of the urdf are located. (default - None)
        - root_joint - Joint at the base of the model (default - None)
        - root_joint_name - Name for the root_joint (default - "root_joint")
        - verbose - print information of parsing (default - False)
        - meshLoader - object used to load meshes (default - hpp::fcl::MeshLoader)
        - geometry_types - Which geometry model to load. Can be pin.GeometryType.COLLISION, pin.GeometryType.VISUAL or both. (default - [pin.GeometryType.COLLISION, pin.GeometryType.VISUAL])
        - mimic - If urdf mimic joints should be parsed or not (default - False)
    Return:
        Tuple of the models, in this order : model, collision model, and visual model.

    Example:
        model, collision_model, visual_model = buildModelsFromUrdf(filename, root_joint, verbose, meshLoader, geometry_types, root_joint_name="root_joint_name")

    Remark: In the URDF format, a joint of type fixed can be defined. For efficiency reasons, it is treated as operational frame and not as a joint of the model.
    """
    # Handle the switch from old to new api
    arg_keys = [
        "package_dirs",
        "root_joint",
        "verbose",
        "meshLoader",
        "geometry_types",
        "mimic",
    ]
    if len(args) >= 3:
        if isinstance(args[2], str):
            arg_keys = [
                "package_dirs",
                "root_joint",
                "root_joint_name",
                "verbose",
                "meshLoader",
                "geometry_types",
                "mimic",
            ]

    for key, arg in zip(arg_keys, args):
        if key in kwargs.keys():
            raise TypeError("Function got multiple values for argument ", key)
        else:
            kwargs[key] = arg

    return _buildModelsFromUrdf(filename, **kwargs)


def _buildModelsFromUrdf(
    filename,
    package_dirs=None,
    root_joint=None,
    root_joint_name=None,
    verbose=False,
    meshLoader=None,
    geometry_types=None,
    mimic=False,
) -> tuple[pin.Model, pin.GeometryModel, pin.GeometryModel]:
    if geometry_types is None:
        geometry_types = [pin.GeometryType.COLLISION, pin.GeometryType.VISUAL]

    if root_joint is None:
        model = pin.buildModelFromUrdf(filename, mimic)
    elif root_joint is not None and root_joint_name is None:
        model = pin.buildModelFromUrdf(filename, root_joint, mimic)
    else:
        model = pin.buildModelFromUrdf(filename, root_joint, root_joint_name, mimic)

    if verbose and not WITH_HPP_FCL and meshLoader is not None:
        print(
            "Info: MeshLoader is ignored. Pinocchio has not been compiled with HPP-FCL."
        )
    if verbose and not WITH_HPP_FCL_BINDINGS and meshLoader is not None:
        print(
            "Info: MeshLoader is ignored. The HPP-FCL Python bindings have not been installed."
        )
    if package_dirs is None:
        package_dirs = []

    lst = [model]

    if not hasattr(geometry_types, "__iter__"):
        geometry_types = [geometry_types]

    for geometry_type in geometry_types:
        if meshLoader is None or (not WITH_HPP_FCL and not WITH_HPP_FCL_BINDINGS):
            geom_model = pin.buildGeomFromUrdf(
                model, filename, geometry_type, package_dirs=package_dirs
            )
        else:
            geom_model = pin.buildGeomFromUrdf(
                model,
                filename,
                geometry_type,
                package_dirs=package_dirs,
                mesh_loader=meshLoader,
            )
        lst.append(geom_model)

    return tuple(lst)


def createDatas(*models):
    """
    Call createData() on each Model or GeometryModel in input and return the results in
    a tuple. If one of the models is None, the corresponding data object in the result
    is also None.
    """
    return tuple([None if model is None else model.createData() for model in models])


def buildModelsFromSdf(
    filename, *args, **kwargs
) -> tuple[pin.Model, pin.GeometryModel, pin.GeometryModel]:
    """Parse the Sdf file given in input and return a Pinocchio Model and a list of Constraint Models, followed by corresponding GeometryModels of types specified by geometry_types, in the same order as listed.
    Arguments:
        - filename - name of the urdf file to load
        - package_dirs - where the meshes of the urdf are located. (default - None)
        - root_joint - Joint at the base of the model (default - None)
        - root_link_name - Name of the body to use as root of the model (default - "")
        - root_joint_name - Name for the root_joint (default - "root_joint")
        - parent_guidance - Joint names which should be preferred for cases where two joints can qualify as parent. The other joint appears in the constraint_model. If empty, joint appearance order in .sdf is taken as default.
        - verbose - print information of parsing (default - False)
        - meshLoader - object used to load meshes (default - hpp::fcl::MeshLoader)
        - geometry_types - Which geometry model to load. Can be pin.GeometryType.COLLISION, pin.GeometryType.VISUAL, both or None. (default - None])
    Return:
        Tuple of the models, in this order : model, collision model, and visual model.

    Example:
        model, collision_model, visual_model = buildModelsFromSdf(filename, root_joint, root_link_name, parent_guidance, verbose, meshLoader, geometry_types, root_joint_name="root_joint_name")
    """

    arg_keys = [
        "package_dirs",
        "root_joint",
        "root_link_name",
        "parent_guidance",
        "verbose",
        "meshLoader",
        "geometry_types",
    ]
    # Handle the switch from old to new api
    if len(args) >= 4:
        if isinstance(args[3], str):
            arg_keys = [
                "package_dirs",
                "root_joint",
                "root_link_name",
                "root_joint_name",
                "parent_guidance",
                "verbose",
                "meshLoader",
                "geometry_types",
            ]

    for key, arg in zip(arg_keys, args):
        if key in kwargs.keys():
            raise TypeError("Function got multiple values for argument ", key)
        else:
            kwargs[key] = arg

    return _buildModelsFromSdf(filename, **kwargs)


def _buildModelsFromSdf(
    filename,
    package_dirs=None,
    root_joint=None,
    root_link_name="",
    root_joint_name=None,
    parent_guidance=None,
    verbose=False,
    meshLoader=None,
    geometry_types=None,
):
    if parent_guidance is None:
        parent_guidance = []

    if geometry_types is None:
        geometry_types = [pin.GeometryType.COLLISION, pin.GeometryType.VISUAL]

    if root_joint is None:
        model, constraint_models = pin.buildModelFromSdf(
            filename, root_link_name, parent_guidance
        )
    elif root_joint is not None and root_joint_name is None:
        model, constraint_models = pin.buildModelFromSdf(
            filename, root_joint, root_link_name, parent_guidance
        )
    else:
        model, constraint_models = pin.buildModelFromSdf(
            filename, root_joint, root_link_name, root_joint_name, parent_guidance
        )
    if verbose and not WITH_HPP_FCL and meshLoader is not None:
        print(
            "Info: MeshLoader is ignored. Pinocchio has not been compiled with HPP-FCL."
        )
    if verbose and not WITH_HPP_FCL_BINDINGS and meshLoader is not None:
        print(
            "Info: MeshLoader is ignored. The HPP-FCL Python bindings have not been installed."
        )
    if package_dirs is None:
        package_dirs = []

    lst = [model, constraint_models]

    if not hasattr(geometry_types, "__iter__"):
        geometry_types = [geometry_types]

    for geometry_type in geometry_types:
        if meshLoader is None or (not WITH_HPP_FCL and not WITH_HPP_FCL_BINDINGS):
            geom_model = pin.buildGeomFromSdf(
                model, filename, geometry_type, root_link_name, package_dirs
            )
        else:
            geom_model = pin.buildGeomFromSdf(
                model, filename, geometry_type, root_link_name, package_dirs, meshLoader
            )
        lst.append(geom_model)

    return tuple(lst)


def buildModelsFromMJCF(filename, *args, **kwargs):
    """Parse the Mjcf file given in input and return a Pinocchio Model followed by corresponding GeometryModels of types specified by geometry_types, in the same order as listed.
    Arguments:
        - filename - name of the urdf file to load
        - package_dirs - where the meshes of the urdf are located. (default - None)
        - root_joint - Joint at the base of the model (default - None)
        - root_joint_name - Name for the root_joint (default - "root_joint")
        - verbose - print information of parsing (default - False)
        - meshLoader - object used to load meshes (default - hpp::fcl::MeshLoader)
        - geometry_types - Which geometry model to load. Can be pin.GeometryType.COLLISION, pin.GeometryType.VISUAL or both. (default - [pin.GeometryType.COLLISION, pin.GeometryType.VISUAL])
        - contacts - Boolean to know if contraint models are wanted (default - False)
    Return:
        Tuple of the models, in this order : model, collision model, and visual model, or  model, constraint_list, collision model, and visual model, if contacts is True.

    Example:
        model, collision_model, visual_model = buildModelsFromMJCF(filename, root_joint, verbose, meshLoader, geometry_types, root_joint_name="root_joint_name")
    """
    # Handle the switch from old to new api
    arg_keys = ["root_joint", "verbose", "meshLoader", "geometry_types", "contacts"]
    if len(args) >= 2:
        if isinstance(args[1], str):
            arg_keys = [
                "root_joint",
                "root_joint_name",
                "verbose",
                "meshLoader",
                "geometry_types",
                "contacts",
            ]

    for key, arg in zip(arg_keys, args):
        if key in kwargs.keys():
            raise TypeError("Function got multiple values for argument ", key)
        else:
            kwargs[key] = arg

    return _buildModelsFromMJCF(filename, **kwargs)


def _buildModelsFromMJCF(
    filename,
    root_joint=None,
    root_joint_name=None,
    verbose=False,
    meshLoader=None,
    geometry_types=None,
    contacts=False,
):
    if geometry_types is None:
        geometry_types = [pin.GeometryType.COLLISION, pin.GeometryType.VISUAL]

    contact_models = []
    if root_joint is None:
        model = pin.buildModelFromMJCF(filename)
    elif root_joint is not None and root_joint_name is None:
        model = pin.buildModelFromMJCF(filename, root_joint)
    else:
        model, contact_models = pin.buildModelFromMJCF(
            filename, root_joint, root_joint_name
        )

    if verbose and not WITH_HPP_FCL and meshLoader is not None:
        print(
            "Info: MeshLoader is ignored. Pinocchio has not been compiled with HPP-FCL."
        )
    if verbose and not WITH_HPP_FCL_BINDINGS and meshLoader is not None:
        print(
            "Info: MeshLoader is ignored. "
            "The HPP-FCL Python bindings have not been installed."
        )

    lst = [model]
    if contacts:
        lst.append(contact_models)

    if not hasattr(geometry_types, "__iter__"):
        geometry_types = [geometry_types]

    for geometry_type in geometry_types:
        if meshLoader is None or (not WITH_HPP_FCL and not WITH_HPP_FCL_BINDINGS):
            geom_model = pin.buildGeomFromMJCF(model, filename, geometry_type)
        else:
            geom_model = pin.buildGeomFromMJCF(
                model, filename, geometry_type, mesh_loader=meshLoader
            )
        lst.append(geom_model)

    return tuple(lst)



import abc
import time
from pathlib import Path

import numpy as np

from .. import pinocchio_pywrap_default as pin
from ..shortcuts import createDatas

try:
    import imageio

    IMAGEIO_SUPPORT = True
except ImportError:
    IMAGEIO_SUPPORT = False


class BaseVisualizer(abc.ABC):
    """
    Pinocchio visualizers are employed to easily display a model at a given
    configuration.
    BaseVisualizer is not meant to be directly employed, but only to provide a uniform
    interface and a few common methods.
    New visualizers should extend this class and override its methods as neeeded.
    """

    _video_writer = None

    _video_writer = None

    def __init__(
        self,
        model=pin.Model(),
        collision_model=None,
        visual_model=None,
        copy_models=False,
        data=None,
        collision_data=None,
        visual_data=None,
    ):
        """
        Construct a display from the given model, collision model, and visual model.
        If copy_models is True, the models are copied. Otherwise, they are simply kept
        as a reference.
        """

        if copy_models:
            self.model = model.copy()
            self.collision_model = collision_model.copy()
            self.visual_model = visual_model.copy()
        else:
            self.model = model
            self.collision_model = collision_model
            self.visual_model = visual_model

        if data is None:
            self.data = self.model.createData()
        else:
            self.data = data

        if collision_data is None and self.collision_model is not None:
            self.collision_data = self.collision_model.createData()
        else:
            self.collision_data = collision_data

        if visual_data is None and self.visual_model is not None:
            self.visual_data = self.visual_model.createData()
        else:
            self.visual_data = visual_data

    def rebuildData(self):
        """Re-build the data objects. Needed if the models were modified.
        Warning: this will delete any information stored in all data objects."""
        self.data, self.collision_data, self.visual_data = createDatas(
            self.model, self.collision_model, self.visual_model
        )

    def getViewerNodeName(self, geometry_object, geometry_type):
        """Return the name of the geometry object inside the viewer."""
        pass

    def initViewer(self, *args, **kwargs):
        """Init the viewer by loading the gui and creating a window."""
        pass

    def loadViewerModel(self, *args, **kwargs):
        """Create the scene displaying the robot meshes in the viewer"""
        pass

    def reload(self, new_geometry_object, geometry_type=None):
        """Reload a geometry_object given by its type"""
        pass

    def clean(self):
        """Delete all the objects from the whole scene"""
        pass

    @abc.abstractmethod
    def display(self, q=None):
        """
        Display the robot at configuration q or refresh the rendering
        from the current placements contained in data by placing all the bodies in the
        viewer.
        """

    @abc.abstractmethod
    def displayCollisions(self, visibility):
        """Set whether to display collision objects or not."""

    @abc.abstractmethod
    def displayVisuals(self, visibility):
        """Set whether to display visual objects or not."""

    @abc.abstractmethod
    def setBackgroundColor(self):
        """Set the visualizer background color."""

    @abc.abstractmethod
    def setCameraTarget(self, target):
        """Set the camera target."""

    @abc.abstractmethod
    def setCameraPosition(self, position: np.ndarray):
        """Set the camera's 3D position."""

    @abc.abstractmethod
    def setCameraZoom(self, zoom: float):
        """Set camera zoom value."""

    @abc.abstractmethod
    def setCameraPose(self, pose: np.ndarray = np.eye(4)):
        """Set camera 6D pose using a 4x4 matrix."""

    @abc.abstractmethod
    def captureImage(self, w=None, h=None):
        """Captures an image from the viewer and returns an RGB array."""

    @abc.abstractmethod
    def disableCameraControl(self):
        """Disable camera manual control"""

    @abc.abstractmethod
    def enableCameraControl(self):
        """Enable camera manual control"""

    @abc.abstractmethod
    def drawFrameVelocities(self, *args, **kwargs):
        """Draw current frame velocities."""

    def sleep(self, dt):
        time.sleep(dt)

    def has_video_writer(self):
        return self._video_writer is not None

    def play(self, q_trajectory, dt=None, callback=None, capture=False, **kwargs):
        """
        Play a trajectory with given time step. Optionally capture RGB images and
        returns them.
        """
        nsteps = len(q_trajectory)
        if not capture:
            capture = self.has_video_writer()

        imgs = []
        for i in range(nsteps):
            t0 = time.time()
            self.display(q_trajectory[i])
            if callback is not None:
                callback(i, **kwargs)
            if capture:
                img_arr = self.captureImage()
                if not self.has_video_writer():
                    imgs.append(img_arr)
                else:
                    self._video_writer.append_data(img_arr)
            t1 = time.time()
            elapsed_time = t1 - t0
            if dt is not None and elapsed_time < dt:
                self.sleep(dt - elapsed_time)
        if capture and not self.has_video_writer():
            return imgs

    def create_video_ctx(self, filename=None, fps=30, directory=None, **kwargs):
        """
        Create a video recording context, generating the output filename if necessary.

        Code inspired from https://github.com/petrikvladimir/RoboMeshCat.
        """
        if not IMAGEIO_SUPPORT:
            import contextlib
            import warnings

            warnings.warn(
                "Video context cannot be created because imageio is not available.",
                UserWarning,
            )
            return contextlib.nullcontext()
        if filename is None:
            if directory is None:
                from tempfile import gettempdir

                directory = gettempdir()
            directory = Path(directory)
            f_fmt = "%Y%m%d_%H%M%S"
            ext = "mp4"
            filename = directory / time.strftime(f"{f_fmt}.{ext}")
        return VideoContext(self, fps, filename)


class VideoContext:
    def __init__(self, viz, fps, filename, **kwargs):
        self.viz = viz
        self.vid_writer = imageio.get_writer(filename, fps=fps, **kwargs)

    def __enter__(self):
        print("[Entering video recording context]")
        self.viz._video_writer = self.vid_writer

    def __exit__(self, *args):
        self.vid_writer.close()
        self.viz._video_writer = None


__all__ = ["BaseVisualizer"]



#!/usr/bin/env python3
#
# SPDX-License-Identifier: BSD-2-Clause

import unittest
from pathlib import Path

import pinocchio as pin


class TestRobotWrapper(unittest.TestCase):
    def setUp(self):
        self.current_dir = Path(__file__).parent

    def test_mjcf_without_root_joint(self):
        model_path = self.current_dir.parent / "models" / "test_mjcf.xml"
        robot = pin.RobotWrapper.BuildFromMJCF(model_path)
        self.assertEqual(robot.nq, 6)
        self.assertEqual(robot.nv, 5)
        self.assertEqual(robot.model.njoints, 4)

    def test_mjcf_with_root_joint(self):
        model_path = self.current_dir.parent / "models" / "test_mjcf.xml"
        robot = pin.RobotWrapper.BuildFromMJCF(model_path, pin.JointModelFreeFlyer())
        self.assertEqual(robot.model.names[1], "root_joint")

    def test_mjcf_with_root_joint_and_root_joint_name(self):
        model_path = self.current_dir.parent / "models" / "test_mjcf.xml"
        name_ = "freeflyer_joint"
        robot = pin.RobotWrapper.BuildFromMJCF(
            model_path, pin.JointModelFreeFlyer(), name_
        )
        self.assertEqual(robot.model.names[1], name_)

    def test_urdf_with_root_joint(self):
        model_path = self.current_dir.parent / "models" / "3DOF_planar.urdf"
        robot = pin.RobotWrapper.BuildFromURDF(
            model_path, [], pin.JointModelFreeFlyer()
        )
        self.assertEqual(robot.model.names[1], "root_joint")

    def test_urdf_with_root_joint_and_root_joint_name(self):
        model_path = self.current_dir.parent / "models" / "3DOF_planar.urdf"
        name_ = "freeflyer_joint"
        robot = pin.RobotWrapper.BuildFromURDF(
            model_path, [], pin.JointModelFreeFlyer(), name_
        )
        self.assertEqual(robot.model.names[1], name_)

    def test_urdf_with_str_pkg_dirs(self):
        model_path = self.current_dir.parent / "models" / "3DOF_planar.urdf"
        package_dir = self.current_dir.parent / "models"
        robot = pin.RobotWrapper.BuildFromURDF(
            model_path, str(package_dir), pin.JointModelFreeFlyer()
        )
        self.assertEqual(robot.model.names[1], "root_joint")

    def test_urdf_with_path_pkg_dirs(self):
        model_path = self.current_dir.parent / "models" / "3DOF_planar.urdf"
        package_dir = self.current_dir.parent / "models"
        robot = pin.RobotWrapper.BuildFromURDF(
            model_path, package_dir, pin.JointModelFreeFlyer()
        )
        self.assertEqual(robot.model.names[1], "root_joint")

    def test_urdf_with_str_list_pkg_dirs(self):
        model_path = self.current_dir.parent / "models" / "3DOF_planar.urdf"
        package_dir = self.current_dir.parent / "models"
        robot = pin.RobotWrapper.BuildFromURDF(
            model_path, [str(package_dir)], pin.JointModelFreeFlyer()
        )
        self.assertEqual(robot.model.names[1], "root_joint")

    def test_urdf_with_path_list_pkg_dirs(self):
        model_path = self.current_dir.parent / "models" / "3DOF_planar.urdf"
        package_dir = self.current_dir.parent / "models"
        robot = pin.RobotWrapper.BuildFromURDF(
            model_path, [package_dir], pin.JointModelFreeFlyer()
        )
        self.assertEqual(robot.model.names[1], "root_joint")

    def test_urdf_with_None_pkg_dirs(self):
        model_path = self.current_dir.parent / "models" / "3DOF_planar.urdf"
        robot = pin.RobotWrapper.BuildFromURDF(
            model_path, None, pin.JointModelFreeFlyer()
        )
        self.assertEqual(robot.model.names[1], "root_joint")

    @unittest.skipUnless(pin.WITH_SDFORMAT, "Needs SDFORMAT")
    def test_sdf_with_root_joint(self):
        model_path = self.current_dir.parent.parent / "models" / "simple_humanoid.sdf"
        mesh_path = self.current_dir.parent.parent / "models"
        robot = pin.RobotWrapper.BuildFromSDF(
            model_path, [mesh_path], pin.JointModelFreeFlyer(), verbose=True
        )
        self.assertEqual(robot.model.names[1], "root_joint")

    @unittest.skipUnless(pin.WITH_SDFORMAT, "Needs SDFORMAT")
    def test_sdf_with_root_joint_and_root_joint_name(self):
        model_path = self.current_dir.parent.parent / "models" / "simple_humanoid.sdf"
        mesh_path = self.current_dir.parent.parent / "models"
        name_ = "freeflyer_joint"
        robot = pin.RobotWrapper.BuildFromSDF(
            model_path, [mesh_path], pin.JointModelFreeFlyer(), root_joint_name=name_
        )
        self.assertEqual(robot.model.names[1], name_)


if __name__ == "__main__":
    unittest.main()



#! /usr/bin/env python3
# Find common dynamics public symbols between shared libraries.

import argparse
import itertools
import pathlib
import subprocess


def generate_symbols(shared_library: pathlib.Path) -> set[str]:
    # Show symbol
    # -D: Dynamic
    # -C: Demangled
    # -U: Defined
    # -W: Non weak
    result = subprocess.run(
        ["nm", "-DCUW", "--format=sysv", str(shared_library)],
        capture_output=True,
        text=True,
    )
    output = result.stdout
    lines_split = (line.split("|") for line in output.splitlines() if "|" in line)
    # Only keep lines with exported (upper case) symbols.
    # `u` is also a global symbol, but if we always build with compatible libraries,
    # there is no issue to find it in many places.
    return set([line[0].strip() for line in lines_split if line[2].strip().isupper()])


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="common_symbol",
        description="Find common dynamics public symbols between shared libraries.",
    )
    parser.add_argument("shared_libraries", nargs="+", type=pathlib.Path)

    args = parser.parse_args()
    symbols = [
        (shared_library, generate_symbols(shared_library))
        for shared_library in args.shared_libraries
    ]

    for lib1, lib2 in itertools.combinations(symbols, 2):
        print(f"Common symbols between {lib1[0]} and {lib2[0]}")
        common_symbols = lib1[1].intersection(lib2[1])
        for common in common_symbols:
            print(f"\t{common}")
        print()



import sys
import time

import numpy as np
import pinocchio as pin
from pinocchio.visualize import MeshcatVisualizer

# This example show how to:
#  - Construct a kinematics chain with ModelGraph API
#  - Add Geometries to a graph
#  - create a geometry model from a graph
#  - visualize it

g = pin.graph.ModelGraph()

# Adding bodies to the pin.graph.
# In pinocchio, since bodies are represented as frames,
# there's 2 ways to add a body to the graph :
#   - addBody, helper function
#   - addFrame, where you create the BodyFrame, yourself
g.addBody("body1", pin.Inertia.Identity())
g.addFrame("body2", pin.graph.BodyFrame(pin.Inertia.Identity()))
g.addFrame("body3", pin.graph.BodyFrame(pin.Inertia.Identity()))

# Adding a sensor to the pin.graph.
g.addFrame("sensor1", pin.graph.SensorFrame())

# Now we add the joints between every body/sensor we have in the pin.graph.
pose_b1_to_j1 = pin.SE3(np.eye(3), np.array([0.1, 0, 0]))  # pose of joint j1 wrt body1
r = np.array([[0, 0, -1], [0, 1, 0], [1, 0, 0]])
pose_j1_to_b2 = pin.SE3(r, np.array([0.4, 0, 0]))  # pose of body2 wrt joint j1

# There are 2 ways to add joints to the graph.
# This helper function, where you specify the basics of a joint only.
g.addJoint(
    "j1",
    pin.graph.JointRevolute(np.array([0.0, 0.0, 1.0])),
    "body1",
    pose_b1_to_j1,
    "body2",
    pose_j1_to_b2,
)

# j2 will be biased by 50cm.

pose_b2_to_j2 = pin.SE3(
    np.eye(3), np.array([0.0, 0.2, -0.4])
)  # pose of joint j2 wrt body2
pose_j2_to_b3 = pin.SE3(np.eye(3), np.array([0, 0, 0]))  # pose of body3 wrt joint j2

# To add a joint with more details, such as limits, bias... the
# builder interface is necessary.
g.edgeBuilder().withName("j2").withJointType(
    pin.graph.JointPrismatic(np.array([0, 0, 1]))
).withSourceVertex("body2").withSourcePose(pose_b2_to_j2).withTargetVertex(
    "body3"
).withTargetPose(pose_j2_to_b3).build()

# Now we can choose which body will be our root its position, and build the model
kinematics_chain_from_body1 = pin.graph.buildModel(g, "body1", pin.SE3.Identity())
print("Kinematics chain from body1:")
print(kinematics_chain_from_body1)

# Now if you want to visualize, it's possible to add geometries to each body.
# Same as for joints, there is a builder interface, where all geometries
# parameters can be redefined.
# Type is to define in which geometry model the geometry will be added,
# it can be either visual, collision or both.
g.geometryBuilder().withBody("body1").withGeomType(pin.graph.GeomType.BOTH).withName(
    "body1_geom1"
).withPlacement(pin.SE3.Identity()).withGeom(
    pin.graph.Box(np.array([0.2, 0.2, 0.2]))
).withColor(np.array([0.3, 0.3, 0.7, 1])).build()
g.geometryBuilder().withBody("body2").withGeomType(pin.graph.GeomType.BOTH).withName(
    "body2_geom1"
).withPlacement(pin.SE3.Identity()).withGeom(
    pin.graph.Cylinder(np.array([0.1, 0.8]))
).withColor(np.array([0.7, 0.3, 0.3, 1])).build()
g.geometryBuilder().withBody("body3").withGeomType(pin.graph.GeomType.BOTH).withName(
    "body3_geom1"
).withPlacement(pin.SE3.Identity()).withGeom(pin.graph.Sphere(0.1)).build()

# And afterward build a pinocchio geometry model
visual_model = pin.graph.buildGeometryModel(g, kinematics_chain_from_body1, pin.VISUAL)
collision_model = pin.graph.buildGeometryModel(
    g, kinematics_chain_from_body1, pin.COLLISION
)

# Start a new MeshCat server and client.
# Note: the server can also be started separately using the "meshcat-server" command in
# a terminal:
# this enables the server to remain active after the current script ends.
#
# Option open=True pens the visualizer.
# Note: the visualizer can also be opened separately by visiting the provided URL.
try:
    vizer = MeshcatVisualizer(
        kinematics_chain_from_body1, collision_model, visual_model
    )
    vizer.initViewer(open=True)
except ImportError as err:
    print(
        "Error while initializing the viewer. "
        "It seems you should install Python meshcat"
    )
    print(err)
    sys.exit(0)

vizer.loadViewerModel()

q = np.array([0.3, 0.4])
vizer.display(q)
time.sleep(10.0)



import unittest

import numpy as np
import pinocchio as pin
from pinocchio.utils import rand, zero


class TestMotionBindings(unittest.TestCase):
    def test_zero_getters(self):
        v = pin.Motion.Zero()
        self.assertTrue(np.allclose(zero(3), v.linear))
        self.assertTrue(np.allclose(zero(3), v.angular))
        self.assertTrue(np.allclose(zero(6), v.vector))

    # TODO: this is not nice, since a random vector can *in theory* be zero
    def test_setRandom(self):
        v = pin.Motion.Zero()
        v.setRandom()
        self.assertFalse(np.allclose(zero(3), v.linear))
        self.assertFalse(np.allclose(zero(3), v.angular))
        self.assertFalse(np.allclose(zero(6), v.vector))

    def test_setZero(self):
        v = pin.Motion.Zero()
        v.setRandom()
        v.setZero()
        self.assertTrue(np.allclose(zero(3), v.linear))
        self.assertTrue(np.allclose(zero(3), v.angular))
        self.assertTrue(np.allclose(zero(6), v.vector))

    def test_set_linear(self):
        v = pin.Motion.Zero()
        lin = rand(3)
        v.linear = lin
        self.assertTrue(np.allclose(v.linear, lin))

        v.linear[1] = 1.0
        self.assertTrue(v.linear[1] == 1.0)

    def test_set_angular(self):
        v = pin.Motion.Zero()
        ang = rand(3)
        v.angular = ang
        self.assertTrue(np.allclose(v.angular, ang))

        v.angular[1] = 1.0
        self.assertTrue(v.angular[1] == 1.0)

    def test_set_vector(self):
        v = pin.Motion.Zero()
        vec = rand(6)
        v.vector = vec
        self.assertTrue(np.allclose(v.vector, vec))

    def test_internal_sums(self):
        v1 = pin.Motion.Random()
        v2 = pin.Motion.Random()
        self.assertTrue(np.allclose((v1 + v2).vector, v1.vector + v2.vector))
        self.assertTrue(np.allclose((v1 - v2).vector, v1.vector - v2.vector))

    def test_se3_action(self):
        m = pin.SE3.Random()
        v = pin.Motion.Random()
        self.assertTrue(np.allclose((m * v).vector, m.action.dot(v.vector)))
        self.assertTrue(np.allclose(m.act(v).vector, m.action.dot(v.vector)))
        self.assertTrue(
            np.allclose((m.actInv(v)).vector, np.linalg.inv(m.action).dot(v.vector))
        )
        self.assertTrue(np.allclose((v ^ v).vector, zero(6)))

    def test_conversion(self):
        m = pin.Motion.Random()
        m_array = np.array(m)

        m_from_array = pin.Motion(m_array)

        self.assertTrue(m_from_array == m)

    def test_several_init(self):
        for _ in range(100000):
            v = pin.Motion.Zero() + pin.Motion.Zero()
            self.assertTrue(np.allclose(v.vector, zero(6)))


if __name__ == "__main__":
    unittest.main()



import os
from pathlib import Path
from sys import argv

import pinocchio

# This path refers to Pinocchio source code but you can define your own directory here.
model_dir = Path(os.environ.get("EXAMPLE_ROBOT_DATA_MODEL_DIR"))

# You should change here to set up your own URDF file or just pass it as an argument of
# this example.
urdf_filename = (
    model_dir / "ur_description/urdf/ur5_robot.urdf" if len(argv) < 2 else argv[1]
)

# Load the urdf model
model = pinocchio.buildModelFromUrdf(urdf_filename)
print("model name: " + model.name)

# Create data required by the algorithms
data = model.createData()

# Sample a random configuration
q = pinocchio.randomConfiguration(model)
print(f"q: {q.T}")

# Perform the forward kinematics over the kinematic tree
pinocchio.forwardKinematics(model, data, q)

# Print out the placement of each joint of the kinematic tree
for name, oMi in zip(model.names, data.oMi):
    print("{:<24} : {: .2f} {: .2f} {: .2f}".format(name, *oMi.translation.T.flat))



import unittest

import numpy as np
import pinocchio as pin
from test_case import PinocchioTestCase as TestCase


class TestRNEA(TestCase):
    def setUp(self):
        self.model = pin.buildSampleModelHumanoidRandom()
        self.data = self.model.createData()

        qmax = np.full((self.model.nq, 1), np.pi)
        self.q = pin.randomConfiguration(self.model, -qmax, qmax)
        self.v = np.random.rand(self.model.nv)
        self.a = np.random.rand(self.model.nv)

        self.fext = []
        for _ in range(self.model.njoints):
            self.fext.append(pin.Force.Random())

    def test_rnea(self):
        model = self.model
        tau = pin.rnea(self.model, self.data, self.q, self.v, self.a)

        null_fext = pin.StdVec_Force()
        for k in range(model.njoints):
            null_fext.append(pin.Force.Zero())

        tau_null_fext = pin.rnea(
            self.model, self.data, self.q, self.v, self.a, null_fext
        )
        self.assertApprox(tau_null_fext, tau)

        null_fext_list = []
        for f in null_fext:
            null_fext_list.append(f)

        print("size:", len(null_fext_list))
        tau_null_fext_list = pin.rnea(
            self.model, self.data, self.q, self.v, self.a, null_fext_list
        )
        self.assertApprox(tau_null_fext_list, tau)

    def test_nle(self):
        model = self.model

        tau = pin.nonLinearEffects(model, self.data, self.q, self.v)

        data2 = model.createData()
        tau_ref = pin.rnea(model, data2, self.q, self.v, self.a * 0)

        self.assertApprox(tau, tau_ref)

    def test_generalized_gravity(self):
        model = self.model

        tau = pin.computeGeneralizedGravity(model, self.data, self.q)

        data2 = model.createData()
        tau_ref = pin.rnea(model, data2, self.q, self.v * 0, self.a * 0)

        self.assertApprox(tau, tau_ref)

    def test_static_torque(self):
        model = self.model

        tau = pin.computeStaticTorque(model, self.data, self.q, self.fext)

        data2 = model.createData()
        tau_ref = pin.rnea(model, data2, self.q, self.v * 0, self.a * 0, self.fext)

        self.assertApprox(tau, tau_ref)

    def test_coriolis_matrix(self):
        model = self.model

        C = pin.computeCoriolisMatrix(model, self.data, self.q, self.v)

        data2 = model.createData()
        tau_coriolis_ref = pin.rnea(
            model, data2, self.q, self.v, self.a * 0
        ) - pin.rnea(model, data2, self.q, self.v * 0, self.a * 0)

        self.assertApprox(tau_coriolis_ref, C.dot(self.v))


if __name__ == "__main__":
    unittest.main()



import unittest
from pathlib import Path

import numpy as np
import pinocchio as pin
from test_case import PinocchioTestCase


class TestFrameBindings(PinocchioTestCase):
    def setUp(self):
        self.model = pin.buildSampleModelHumanoidRandom()
        self.parent_idx = (
            self.model.getJointId("rarm2_joint")
            if self.model.existJointName("rarm2_joint")
            else (self.model.njoints - 1)
        )
        self.frame_name = self.model.names[self.parent_idx] + "_frame"
        self.frame_placement = pin.SE3.Random()
        self.frame_type = pin.FrameType.OP_FRAME
        self.model.addFrame(
            pin.Frame(
                self.frame_name,
                self.parent_idx,
                0,
                self.frame_placement,
                self.frame_type,
            )
        )
        self.frame_idx = self.model.getFrameId(self.frame_name)

    def tearDown(self):
        del self.model

    def test_type_get_set(self):
        f = self.model.frames[self.frame_idx]
        self.assertTrue(f.type == self.frame_type)
        f.type = pin.FrameType.BODY
        self.assertTrue(f.type == pin.FrameType.BODY)

    def test_name_get_set(self):
        f = self.model.frames[self.frame_idx]
        self.assertTrue(f.name == self.frame_name)
        f.name = "new_hip_frame"
        self.assertTrue(f.name == "new_hip_frame")

    def test_parent_get_set(self):
        f = self.model.frames[self.frame_idx]
        self.assertTrue(f.parentJoint == self.parent_idx)
        newparent = self.parent_idx - 1
        f.parentJoint = newparent
        self.assertTrue(f.parentJoint == newparent)

    def test_placement_get_set(self):
        f = self.model.frames[self.frame_idx]
        self.assertTrue(
            np.allclose(f.placement.homogeneous, self.frame_placement.homogeneous)
        )
        new_placement = pin.SE3.Random()
        f.placement = new_placement
        self.assertTrue(np.allclose(f.placement.homogeneous, new_placement.homogeneous))

    def test_frame_equality(self):
        M = pin.SE3.Random()
        frame1 = pin.Frame("name", 1, 2, M, pin.OP_FRAME)
        frame2 = pin.Frame("name", 1, 2, M, pin.OP_FRAME)
        frame3 = pin.Frame("othername", 3, 4, pin.SE3.Random(), pin.BODY)

        self.assertTrue(frame1 == frame2)
        self.assertFalse(frame1 != frame2)
        self.assertTrue(frame1 != frame3)
        self.assertFalse(frame1 == frame3)

    def test_pickle(self):
        import pickle

        frame = pin.Frame("name", 1, 2, pin.SE3.Random(), pin.OP_FRAME)
        filename = Path("frame.pickle")
        with filename.open("wb") as f:
            pickle.dump(frame, f)

        with filename.open("rb") as f:
            frame_copy = pickle.load(f)

        self.assertEqual(frame, frame_copy)

    def test_getters(self):
        data = self.model.createData()
        q = pin.randomConfiguration(self.model)
        v = np.random.rand(self.model.nv)
        a = np.random.rand(self.model.nv)
        pin.forwardKinematics(self.model, data, q, v, a)

        T = pin.updateFramePlacement(self.model, data, self.frame_idx)
        self.assertApprox(T, data.oMi[self.parent_idx].act(self.frame_placement))

        v = pin.getFrameVelocity(self.model, data, self.frame_idx)
        self.assertApprox(v, self.frame_placement.actInv(data.v[self.parent_idx]))
        v = pin.getFrameVelocity(
            self.model, data, self.frame_idx, pin.ReferenceFrame.LOCAL
        )
        self.assertApprox(v, self.frame_placement.actInv(data.v[self.parent_idx]))
        v = pin.getFrameVelocity(
            self.model, data, self.frame_idx, pin.ReferenceFrame.WORLD
        )
        self.assertApprox(v, data.oMi[self.parent_idx].act(data.v[self.parent_idx]))
        v = pin.getFrameVelocity(
            self.model, data, self.frame_idx, pin.ReferenceFrame.LOCAL_WORLD_ALIGNED
        )
        self.assertApprox(
            v,
            pin.SE3(T.rotation, np.zeros(3)).act(
                self.frame_placement.actInv(data.v[self.parent_idx])
            ),
        )

        a = pin.getFrameAcceleration(self.model, data, self.frame_idx)
        self.assertApprox(a, self.frame_placement.actInv(data.a[self.parent_idx]))
        a = pin.getFrameAcceleration(
            self.model, data, self.frame_idx, pin.ReferenceFrame.LOCAL
        )
        self.assertApprox(a, self.frame_placement.actInv(data.a[self.parent_idx]))
        a = pin.getFrameAcceleration(
            self.model, data, self.frame_idx, pin.ReferenceFrame.WORLD
        )
        self.assertApprox(a, data.oMi[self.parent_idx].act(data.a[self.parent_idx]))
        a = pin.getFrameAcceleration(
            self.model, data, self.frame_idx, pin.ReferenceFrame.LOCAL_WORLD_ALIGNED
        )
        self.assertApprox(
            a,
            pin.SE3(T.rotation, np.zeros(3)).act(
                self.frame_placement.actInv(data.a[self.parent_idx])
            ),
        )

        a = pin.getFrameClassicalAcceleration(self.model, data, self.frame_idx)
        a = pin.getFrameClassicalAcceleration(
            self.model, data, self.frame_idx, pin.ReferenceFrame.LOCAL
        )
        a = pin.getFrameClassicalAcceleration(
            self.model, data, self.frame_idx, pin.ReferenceFrame.WORLD
        )
        a = pin.getFrameClassicalAcceleration(
            self.model, data, self.frame_idx, pin.ReferenceFrame.LOCAL_WORLD_ALIGNED
        )

    def test_frame_algo(self):
        model = self.model
        data = model.createData()

        q = pin.neutral(model)
        v = np.random.rand(model.nv)
        frame_id = self.frame_idx

        J1 = pin.computeFrameJacobian(model, data, q, frame_id)
        J2 = pin.computeFrameJacobian(model, data, q, frame_id, pin.LOCAL)

        self.assertApprox(J1, J2)
        data2 = model.createData()

        pin.computeJointJacobians(model, data2, q)
        J3 = pin.getFrameJacobian(model, data2, frame_id, pin.LOCAL)
        self.assertApprox(J1, J3)

        dJ1 = pin.frameJacobianTimeVariation(model, data, q, v, frame_id, pin.LOCAL)

        data3 = model.createData()
        pin.computeJointJacobiansTimeVariation(model, data3, q, v)

        dJ2 = pin.getFrameJacobianTimeVariation(model, data3, frame_id, pin.LOCAL)
        self.assertApprox(dJ1, dJ2)


if __name__ == "__main__":
    unittest.main()



"""
Load 4 times the UR5 model, plus a plate object on top of them, to feature a simple
parallel robot.
No optimization, this file is just an example of how to load the models.
"""

from pathlib import Path

import numpy as np
from pinocchio import SE3
from pinocchio.robot_wrapper import RobotWrapper
from pinocchio.utils import eye, rotate, se3ToXYZQUATtuple, urdf, zero

PKG = Path("/opt/openrobots/share")
URDF = PKG / "ur5_description/urdf/ur5_gripper.urdf"


def loadRobot(M0, name):
    """
    This function load a UR5 robot n a new model, move the basis to placement <M0>
    and add the corresponding visuals in gepetto viewer with name prefix given by string
    <name>.
    It returns the robot wrapper (model,data).
    """
    robot = RobotWrapper(urdf, [PKG])
    robot.model.jointPlacements[1] = M0 * robot.model.jointPlacements[1]
    robot.visual_model.geometryObjects[0].placement = (
        M0 * robot.visual_model.geometryObjects[0].placement
    )
    robot.visual_data.oMg[0] = M0 * robot.visual_data.oMg[0]
    robot.initDisplay(loadModel=True, viewerRootNodeName="world/" + name)
    return robot


robots = []
# Load 4 Ur5 robots, placed at 0.3m from origin in the 4 directions x,y,-x,-y.
Mt = SE3(eye(3), np.array([0.3, 0, 0]))  # First robot is simply translated
for i in range(4):
    robots.append(loadRobot(SE3(rotate("z", np.pi / 2 * i), zero(3)) * Mt, f"robot{i}"))

# Set up the robots configuration with end effector pointed upward.
q0 = np.array([np.pi / 4, -np.pi / 4, -np.pi / 2, np.pi / 4, np.pi / 2, 0])
for i in range(4):
    robots[i].display(q0)

# Add a new object featuring the parallel robot tool plate.
gepettoViewer = robots[0].viewer.gui
w, h, d = 0.25, 0.25, 0.005
color = [red, green, blue, transparency] = [1, 1, 0.78, 1.0]
gepettoViewer.addBox("world/toolplate", w, h, d, color)
Mtool = SE3(rotate("z", 1.268), np.array([0, 0, 0.77]))
gepettoViewer.applyConfiguration("world/toolplate", se3ToXYZQUATtuple(Mtool))
gepettoViewer.refresh()
